import numpy as np

import matchzoo as mz


def print_deprecation_warning(instance):
    name = instance.__class__.__name__
    print(f"WARNING: {name} will be deprecated in MatchZoo v2.2. "
          "Use `DataGenerator` with callbacks instead.")


class HistogramDataGenerator(mz.DataGenerator):
    def __init__(
        self,
        data_pack: mz.DataPack,
        embedding_matrix: np.ndarray,
        bin_size: int = 30,
        hist_mode: str = 'CH',
        batch_size: int = 32,
        shuffle: bool = True
    ):
        super().__init__(
            data_pack=data_pack,
            batch_size=batch_size,
            shuffle=shuffle,
            callbacks=[
                mz.data_generator.callbacks.Histogram(
                    embedding_matrix=embedding_matrix,
                    bin_size=bin_size,
                    hist_mode=hist_mode
                )
            ]
        )
        print_deprecation_warning(self)


class HistogramPairDataGenerator(mz.DataGenerator):
    def __init__(
        self,
        data_pack: mz.DataPack,
        embedding_matrix: np.ndarray,
        bin_size: int = 30,
        hist_mode: str = 'CH',
        num_dup: int = 1,
        num_neg: int = 1,
        batch_size: int = 32,
        shuffle: bool = True
    ):
        super().__init__(
            data_pack=data_pack,
            mode='pair',
            num_dup=num_dup,
            num_neg=num_neg,
            batch_size=batch_size,
            shuffle=shuffle,
            callbacks=[
                mz.data_generator.callbacks.Histogram(
                    embedding_matrix=embedding_matrix,
                    bin_size=bin_size,
                    hist_mode=hist_mode
                )
            ]
        )
        print_deprecation_warning(self)


class DPoolDataGenerator(mz.DataGenerator):
    def __init__(
        self,
        data_pack: mz.DataPack,
        fixed_length_left: int,
        fixed_length_right: int,
        compress_ratio_left: float = 1,
        compress_ratio_right: float = 1,
        batch_size: int = 32,
        shuffle: bool = True
    ):
        super().__init__(
            data_pack=data_pack,
            shuffle=shuffle,
            batch_size=batch_size,
            callbacks=[
                mz.data_generator.callbacks.DynamicPooling(
                    fixed_length_left=fixed_length_left,
                    fixed_length_right=fixed_length_right,
                    compress_ratio_left=compress_ratio_left,
                    compress_ratio_right=compress_ratio_right
                )
            ]
        )
        print_deprecation_warning(self)


class DPoolPairDataGenerator(mz.DataGenerator):
    def __init__(
        self,
        data_pack: mz.DataPack,
        fixed_length_left: int,
        fixed_length_right: int,
        compress_ratio_left: float = 1,
        compress_ratio_right: float = 1,
        num_dup: int = 1,
        num_neg: int = 1,
        batch_size: int = 32,
        shuffle: bool = True
    ):
        super().__init__(
            data_pack=data_pack,
            mode='pair',
            num_dup=num_dup,
            num_neg=num_neg,
            batch_size=batch_size,
            shuffle=shuffle,
            callbacks=[
                mz.data_generator.callbacks.DynamicPooling(
                    fixed_length_left=fixed_length_left,
                    fixed_length_right=fixed_length_right,
                    compress_ratio_left=compress_ratio_left,
                    compress_ratio_right=compress_ratio_right
                )
            ]
        )
        print_deprecation_warning(self)


class PairDataGenerator(mz.DataGenerator):
    def __init__(
        self,
        data_pack: mz.DataPack,
        num_dup: int = 1,
        num_neg: int = 1,
        batch_size: int = 32,
        shuffle: bool = True
    ):
        super().__init__(
            data_pack=data_pack,
            mode='pair',
            num_dup=num_dup,
            num_neg=num_neg,
            batch_size=batch_size,
            shuffle=shuffle,

        )
        print_deprecation_warning(self)


class DynamicDataGenerator(mz.DataGenerator):
    def __init__(self, func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        callback = mz.data_generator.callbacks.LambdaCallback(
            on_batch_data_pack=func)
        self.callbacks.append(callback)
        print_deprecation_warning(self)



"""Basic Preprocessor."""

from tqdm import tqdm

from . import units
from matchzoo import DataPack
from matchzoo.engine.base_preprocessor import BasePreprocessor
from .build_vocab_unit import build_vocab_unit
from .build_unit_from_data_pack import build_unit_from_data_pack
from .chain_transform import chain_transform

tqdm.pandas()


class BasicPreprocessor(BasePreprocessor):
    """
    Baisc preprocessor helper.

    :param fixed_length_left: Integer, maximize length of :attr:`left` in the
        data_pack.
    :param fixed_length_right: Integer, maximize length of :attr:`right` in the
        data_pack.
    :param filter_mode: String, mode used by :class:`FrequenceFilterUnit`, Can
        be 'df', 'cf', and 'idf'.
    :param filter_low_freq: Float, lower bound value used by
        :class:`FrequenceFilterUnit`.
    :param filter_high_freq: Float, upper bound value used by
        :class:`FrequenceFilterUnit`.
    :param remove_stop_words: Bool, use :class:`StopRemovalUnit` unit or not.

    Example:
        >>> import matchzoo as mz
        >>> train_data = mz.datasets.toy.load_data('train')
        >>> test_data = mz.datasets.toy.load_data('test')
        >>> preprocessor = mz.preprocessors.BasicPreprocessor(
        ...     fixed_length_left=10,
        ...     fixed_length_right=20,
        ...     filter_mode='df',
        ...     filter_low_freq=2,
        ...     filter_high_freq=1000,
        ...     remove_stop_words=True
        ... )
        >>> preprocessor = preprocessor.fit(train_data, verbose=0)
        >>> preprocessor.context['input_shapes']
        [(10,), (20,)]
        >>> preprocessor.context['vocab_size']
        225
        >>> processed_train_data = preprocessor.transform(train_data,
        ...                                               verbose=0)
        >>> type(processed_train_data)
        <class 'matchzoo.data_pack.data_pack.DataPack'>
        >>> test_data_transformed = preprocessor.transform(test_data,
        ...                                                verbose=0)
        >>> type(test_data_transformed)
        <class 'matchzoo.data_pack.data_pack.DataPack'>

    """

    def __init__(self, fixed_length_left: int = 30,
                 fixed_length_right: int = 30,
                 filter_mode: str = 'df',
                 filter_low_freq: float = 2,
                 filter_high_freq: float = float('inf'),
                 remove_stop_words: bool = False):
        """Initialization."""
        super().__init__()
        self._fixed_length_left = fixed_length_left
        self._fixed_length_right = fixed_length_right
        self._left_fixedlength_unit = units.FixedLength(
            self._fixed_length_left,
            pad_mode='post'
        )
        self._right_fixedlength_unit = units.FixedLength(
            self._fixed_length_right,
            pad_mode='post'
        )
        self._filter_unit = units.FrequencyFilter(
            low=filter_low_freq,
            high=filter_high_freq,
            mode=filter_mode
        )
        self._units = self._default_units()
        if remove_stop_words:
            self._units.append(units.stop_removal.StopRemoval())

    def fit(self, data_pack: DataPack, verbose: int = 1):
        """
        Fit pre-processing context for transformation.

        :param data_pack: data_pack to be preprocessed.
        :param verbose: Verbosity.
        :return: class:`BasicPreprocessor` instance.
        """
        data_pack = data_pack.apply_on_text(chain_transform(self._units),
                                            verbose=verbose)
        fitted_filter_unit = build_unit_from_data_pack(self._filter_unit,
                                                       data_pack,
                                                       flatten=False,
                                                       mode='right',
                                                       verbose=verbose)
        data_pack = data_pack.apply_on_text(fitted_filter_unit.transform,
                                            mode='right', verbose=verbose)
        self._context['filter_unit'] = fitted_filter_unit

        vocab_unit = build_vocab_unit(data_pack, verbose=verbose)
        self._context['vocab_unit'] = vocab_unit

        vocab_size = len(vocab_unit.state['term_index']) + 1
        self._context['vocab_size'] = vocab_size
        self._context['embedding_input_dim'] = vocab_size
        self._context['input_shapes'] = [(self._fixed_length_left,),
                                         (self._fixed_length_right,)]

        return self

    def transform(self, data_pack: DataPack, verbose: int = 1) -> DataPack:
        """
        Apply transformation on data, create fixed length representation.

        :param data_pack: Inputs to be preprocessed.
        :param verbose: Verbosity.

        :return: Transformed data as :class:`DataPack` object.
        """
        data_pack = data_pack.copy()
        data_pack.apply_on_text(chain_transform(self._units), inplace=True,
                                verbose=verbose)

        data_pack.apply_on_text(self._context['filter_unit'].transform,
                                mode='right', inplace=True, verbose=verbose)
        data_pack.apply_on_text(self._context['vocab_unit'].transform,
                                mode='both', inplace=True, verbose=verbose)
        data_pack.append_text_length(inplace=True, verbose=verbose)
        data_pack.apply_on_text(self._left_fixedlength_unit.transform,
                                mode='left', inplace=True, verbose=verbose)
        data_pack.apply_on_text(self._right_fixedlength_unit.transform,
                                mode='right', inplace=True, verbose=verbose)

        max_len_left = self._fixed_length_left
        max_len_right = self._fixed_length_right

        data_pack.left['length_left'] = \
            data_pack.left['length_left'].apply(
                lambda val: min(val, max_len_left))

        data_pack.right['length_right'] = \
            data_pack.right['length_right'].apply(
                lambda val: min(val, max_len_right))
        return data_pack



import typing

import numpy as np

from .unit import Unit


class FixedLength(Unit):
    """
    FixedLengthUnit Class.

    Process unit to get the fixed length text.

    Examples:
        >>> from matchzoo.preprocessors.units import FixedLength
        >>> fixedlen = FixedLength(3)
        >>> fixedlen.transform(list(range(1, 6))) == [3, 4, 5]
        True
        >>> fixedlen.transform(list(range(1, 3))) == [0, 1, 2]
        True

    """

    def __init__(
        self,
        text_length: int,
        pad_value: typing.Union[int, str] = 0,
        pad_mode: str = 'pre',
        truncate_mode: str = 'pre'
    ):
        """
        Class initialization.

        :param text_length: fixed length of the text.
        :param pad_value: if text length is smaller than :attr:`text_length`,
            filling text with :attr:`pad_value`.
        :param pad_mode: String, `pre` or `post`:
            pad either before or after each sequence.
        :param truncate_mode: String, `pre` or `post`:
            remove values from sequences larger than :attr:`text_length`,
            either at the beginning or at the end of the sequences.
        """
        self._text_length = text_length
        self._pad_value = pad_value
        self._pad_mode = pad_mode
        self._truncate_mode = truncate_mode

    def transform(self, input_: list) -> list:
        """
        Transform list of tokenized tokens into the fixed length text.

        :param input_: list of tokenized tokens.

        :return tokens: list of tokenized tokens in fixed length.
        """
        # padding process can not handle empty list as input
        if len(input_) == 0:
            input_ = [self._pad_value]
        np_tokens = np.array(input_)
        fixed_tokens = np.full([self._text_length], self._pad_value,
                               dtype=np_tokens.dtype)

        if self._truncate_mode == 'pre':
            trunc_tokens = input_[-self._text_length:]
        elif self._truncate_mode == 'post':
            trunc_tokens = input_[:self._text_length]
        else:
            raise ValueError('{} is not a vaild '
                             'truncate mode.'.format(self._truncate_mode))

        if self._pad_mode == 'post':
            fixed_tokens[:len(trunc_tokens)] = trunc_tokens
        elif self._pad_mode == 'pre':
            fixed_tokens[-len(trunc_tokens):] = trunc_tokens
        else:
            raise ValueError('{} is not a vaild '
                             'pad mode.'.format(self._pad_mode))

        return fixed_tokens.tolist()



"""An implementation of aNMM Model."""

import keras
from keras.activations import softmax
from keras.initializers import RandomUniform

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class ANMM(BaseModel):
    """
    ANMM Model.

    Examples:
        >>> model = ANMM()
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(with_embedding=True)
        params.add(Param(
            name='dropout_rate', value=0.1,
            desc="The dropout rate.",
            hyper_space=hyper_spaces.quniform(0, 1, 0.05)
        ))
        params.add(Param(
            name='num_layers', value=2,
            desc="Number of hidden layers in the MLP "
                 "layer."
        ))
        params.add(Param(
            name='hidden_sizes', value=[30, 30],
            desc="Number of hidden size for each hidden"
                 " layer"
        ))
        return params

    def build(self):
        """
        Build model structure.

        aNMM model based on bin weighting and query term attentions
        """
        # query is [batch_size, left_text_len]
        # doc is [batch_size, right_text_len, bin_num]
        query, doc = self._make_inputs()
        embedding = self._make_embedding_layer()

        q_embed = embedding(query)
        q_attention = keras.layers.Dense(
            1, kernel_initializer=RandomUniform(), use_bias=False)(q_embed)
        q_text_len = self._params['input_shapes'][0][0]

        q_attention = keras.layers.Lambda(
            lambda x: softmax(x, axis=1),
            output_shape=(q_text_len,)
        )(q_attention)
        d_bin = keras.layers.Dropout(
            rate=self._params['dropout_rate'])(doc)
        for layer_id in range(self._params['num_layers'] - 1):
            d_bin = keras.layers.Dense(
                self._params['hidden_sizes'][layer_id],
                kernel_initializer=RandomUniform())(d_bin)
            d_bin = keras.layers.Activation('tanh')(d_bin)
        d_bin = keras.layers.Dense(
            self._params['hidden_sizes'][self._params['num_layers'] - 1])(
            d_bin)
        d_bin = keras.layers.Reshape((q_text_len,))(d_bin)
        q_attention = keras.layers.Reshape((q_text_len,))(q_attention)
        score = keras.layers.Dot(axes=[1, 1])([d_bin, q_attention])
        x_out = self._make_output_layer()(score)
        self._backend = keras.Model(inputs=[query, doc], outputs=x_out)



"""KNRM model."""
import keras
import keras.backend as K

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class KNRM(BaseModel):
    """
    KNRM model.

    Examples:
        >>> model = KNRM()
        >>> model.params['embedding_input_dim'] =  10000
        >>> model.params['embedding_output_dim'] =  10
        >>> model.params['embedding_trainable'] = True
        >>> model.params['kernel_num'] = 11
        >>> model.params['sigma'] = 0.1
        >>> model.params['exact_sigma'] = 0.001
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls):
        """Get default parameters."""
        params = super().get_default_params(with_embedding=True)
        params.add(Param(
            name='kernel_num',
            value=11,
            hyper_space=hyper_spaces.quniform(low=5, high=20),
            desc="The number of RBF kernels."
        ))
        params.add(Param(
            name='sigma',
            value=0.1,
            hyper_space=hyper_spaces.quniform(
                low=0.01, high=0.2, q=0.01),
            desc="The `sigma` defines the kernel width."
        ))
        params.add(Param(
            name='exact_sigma', value=0.001,
            desc="The `exact_sigma` denotes the `sigma` "
                 "for exact match."
        ))
        return params

    def build(self):
        """Build model."""
        query, doc = self._make_inputs()

        embedding = self._make_embedding_layer()
        q_embed = embedding(query)
        d_embed = embedding(doc)

        mm = keras.layers.Dot(axes=[2, 2], normalize=True)([q_embed, d_embed])

        KM = []
        for i in range(self._params['kernel_num']):
            mu = 1. / (self._params['kernel_num'] - 1) + (2. * i) / (
                self._params['kernel_num'] - 1) - 1.0
            sigma = self._params['sigma']
            if mu > 1.0:
                sigma = self._params['exact_sigma']
                mu = 1.0
            mm_exp = self._kernel_layer(mu, sigma)(mm)
            mm_doc_sum = keras.layers.Lambda(
                lambda x: K.tf.reduce_sum(x, 2))(mm_exp)
            mm_log = keras.layers.Activation(K.tf.log1p)(mm_doc_sum)
            mm_sum = keras.layers.Lambda(
                lambda x: K.tf.reduce_sum(x, 1))(mm_log)
            KM.append(mm_sum)

        phi = keras.layers.Lambda(lambda x: K.tf.stack(x, 1))(KM)
        out = self._make_output_layer()(phi)
        self._backend = keras.Model(inputs=[query, doc], outputs=[out])

    @classmethod
    def _kernel_layer(cls, mu: float, sigma: float) -> keras.layers.Layer:
        """
        Gaussian kernel layer in KNRM.

        :param mu: Float, mean of the kernel.
        :param sigma: Float, sigma of the kernel.
        :return: `keras.layers.Layer`.
        """

        def kernel(x):
            return K.tf.exp(-0.5 * (x - mu) * (x - mu) / sigma / sigma)

        return keras.layers.Activation(kernel)



from .unit import Unit


class DigitRemoval(Unit):
    """Process unit to remove digits."""

    def transform(self, input_: list) -> list:
        """
        Remove digits from list of tokens.

        :param input_: list of tokens to be filtered.

        :return tokens: tokens of tokens without digits.
        """
        return [token for token in input_ if not token.isdigit()]



"""An implementation of Matching Layer."""
import typing

from keras import backend as K
from keras.engine import Layer


class MatchingLayer(Layer):
    """
    Layer that computes a matching matrix between samples in two tensors.

    :param normalize: Whether to L2-normalize samples along the
        dot product axis before taking the dot product.
        If set to True, then the output of the dot product
        is the cosine proximity between the two samples.
    :param matching_type: the similarity function for matching
    :param kwargs: Standard layer keyword arguments.

    Examples:
        >>> import matchzoo as mz
        >>> layer = mz.layers.MatchingLayer(matching_type='dot',
        ...                                 normalize=True)
        >>> num_batch, left_len, right_len, num_dim = 5, 3, 2, 10
        >>> layer.build([[num_batch, left_len, num_dim],
        ...              [num_batch, right_len, num_dim]])

    """

    def __init__(self, normalize: bool = False,
                 matching_type: str = 'dot', **kwargs):
        """:class:`MatchingLayer` constructor."""
        super().__init__(**kwargs)
        self._normalize = normalize
        self._validate_matching_type(matching_type)
        self._matching_type = matching_type
        self._shape1 = None
        self._shape2 = None

    @classmethod
    def _validate_matching_type(cls, matching_type: str = 'dot'):
        valid_matching_type = ['dot', 'mul', 'plus', 'minus', 'concat']
        if matching_type not in valid_matching_type:
            raise ValueError(f"{matching_type} is not a valid matching type, "
                             f"{valid_matching_type} expected.")

    def build(self, input_shape: list):
        """
        Build the layer.

        :param input_shape: the shapes of the input tensors,
            for MatchingLayer we need tow input tensors.
        """
        # Used purely for shape validation.
        if not isinstance(input_shape, list) or len(input_shape) != 2:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on a list of 2 inputs.')
        self._shape1 = input_shape[0]
        self._shape2 = input_shape[1]
        for idx in 0, 2:
            if self._shape1[idx] != self._shape2[idx]:
                raise ValueError(
                    'Incompatible dimensions: '
                    f'{self._shape1[idx]} != {self._shape2[idx]}.'
                    f'Layer shapes: {self._shape1}, {self._shape2}.'
                )

    def call(self, inputs: list, **kwargs) -> typing.Any:
        """
        The computation logic of MatchingLayer.

        :param inputs: two input tensors.
        """
        x1 = inputs[0]
        x2 = inputs[1]
        if self._matching_type == 'dot':
            if self._normalize:
                x1 = K.l2_normalize(x1, axis=2)
                x2 = K.l2_normalize(x2, axis=2)
            return K.tf.expand_dims(K.tf.einsum('abd,acd->abc', x1, x2), 3)
        else:
            if self._matching_type == 'mul':
                def func(x, y):
                    return x * y
            elif self._matching_type == 'plus':
                def func(x, y):
                    return x + y
            elif self._matching_type == 'minus':
                def func(x, y):
                    return x - y
            elif self._matching_type == 'concat':
                def func(x, y):
                    return K.tf.concat([x, y], axis=3)
            else:
                raise ValueError(f"Invalid matching type."
                                 f"{self._matching_type} received."
                                 f"Mut be in `dot`, `mul`, `plus`, "
                                 f"`minus` and `concat`.")
            x1_exp = K.tf.stack([x1] * self._shape2[1], 2)
            x2_exp = K.tf.stack([x2] * self._shape1[1], 1)
            return func(x1_exp, x2_exp)

    def compute_output_shape(self, input_shape: list) -> tuple:
        """
        Calculate the layer output shape.

        :param input_shape: the shapes of the input tensors,
            for MatchingLayer we need tow input tensors.
        """
        if not isinstance(input_shape, list) or len(input_shape) != 2:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on a list of 2 inputs.')
        shape1 = list(input_shape[0])
        shape2 = list(input_shape[1])
        if len(shape1) != 3 or len(shape2) != 3:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on 2 inputs with 3 dimensions.')
        if shape1[0] != shape2[0] or shape1[2] != shape2[2]:
            raise ValueError('A `MatchingLayer` layer should be called '
                             'on 2 inputs with same 0,2 dimensions.')

        if self._matching_type in ['mul', 'plus', 'minus']:
            return shape1[0], shape1[1], shape2[1], shape1[2]
        elif self._matching_type == 'dot':
            return shape1[0], shape1[1], shape2[1], 1
        elif self._matching_type == 'concat':
            return shape1[0], shape1[1], shape2[1], shape1[2] + shape2[2]
        else:
            raise ValueError(f"Invalid `matching_type`."
                             f"{self._matching_type} received."
                             f"Must be in `mul`, `plus`, `minus` "
                             f"`dot` and `concat`.")

    def get_config(self) -> dict:
        """Get the config dict of MatchingLayer."""
        config = {
            'normalize': self._normalize,
            'matching_type': self._matching_type,
        }
        base_config = super(MatchingLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""An implementation of MVLSTM Model."""
import typing

import keras
import keras.backend as K

import matchzoo
from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class MVLSTM(BaseModel):
    """
    MVLSTM Model.

    Examples:
        >>> model = MVLSTM()
        >>> model.params['lstm_units'] = 32
        >>> model.params['top_k'] = 50
        >>> model.params['mlp_num_layers'] = 2
        >>> model.params['mlp_num_units'] = 20
        >>> model.params['mlp_num_fan_out'] = 10
        >>> model.params['mlp_activation_func'] = 'relu'
        >>> model.params['dropout_rate'] = 0.5
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(
            with_embedding=True, with_multi_layer_perceptron=True)
        params.add(Param(name='lstm_units', value=32,
                         desc="Integer, the hidden size in the "
                              "bi-directional LSTM layer."))
        params.add(Param(name='dropout_rate', value=0.0,
                         desc="Float, the dropout rate."))
        params.add(Param(
            'top_k', value=10,
            hyper_space=hyper_spaces.quniform(low=2, high=100),
            desc="Integer, the size of top-k pooling layer."
        ))
        params['optimizer'] = 'adam'
        return params

    def build(self):
        """Build model structure."""
        query, doc = self._make_inputs()

        # Embedding layer
        embedding = self._make_embedding_layer()
        embed_query = embedding(query)
        embed_doc = embedding(doc)

        # Bi-directional LSTM layer
        rep_query = keras.layers.Bidirectional(keras.layers.LSTM(
            self._params['lstm_units'],
            return_sequences=True,
            dropout=self._params['dropout_rate']
        ))(embed_query)
        rep_doc = keras.layers.Bidirectional(keras.layers.LSTM(
            self._params['lstm_units'],
            return_sequences=True,
            dropout=self._params['dropout_rate']
        ))(embed_doc)

        # Top-k matching layer
        matching_matrix = keras.layers.Dot(
            axes=[2, 2], normalize=False)([rep_query, rep_doc])
        matching_signals = keras.layers.Reshape((-1,))(matching_matrix)
        matching_topk = keras.layers.Lambda(
            lambda x: K.tf.nn.top_k(x, k=self._params['top_k'], sorted=True)[0]
        )(matching_signals)

        # Multilayer perceptron layer.
        mlp = self._make_multi_layer_perceptron_layer()(matching_topk)
        mlp = keras.layers.Dropout(
            rate=self._params['dropout_rate'])(mlp)

        x_out = self._make_output_layer()(mlp)
        self._backend = keras.Model(inputs=[query, doc], outputs=x_out)



"""Matchzoo toolkit for token embedding."""

import csv
import typing

import numpy as np
import pandas as pd

import matchzoo as mz


class Embedding(object):
    """
    Embedding class.

    Examples::
        >>> import matchzoo as mz
        >>> train_raw = mz.datasets.toy.load_data()
        >>> pp = mz.preprocessors.NaivePreprocessor()
        >>> train = pp.fit_transform(train_raw, verbose=0)
        >>> vocab_unit = mz.build_vocab_unit(train, verbose=0)
        >>> term_index = vocab_unit.state['term_index']
        >>> embed_path = mz.datasets.embeddings.EMBED_RANK

    To load from a file:
        >>> embedding = mz.embedding.load_from_file(embed_path)
        >>> matrix = embedding.build_matrix(term_index)
        >>> matrix.shape[0] == len(term_index) + 1
        True

    To build your own:
        >>> data = pd.DataFrame(data=[[0, 1], [2, 3]], index=['A', 'B'])
        >>> embedding = mz.Embedding(data)
        >>> matrix = embedding.build_matrix({'A': 2, 'B': 1})
        >>> matrix.shape == (3, 2)
        True

    """

    def __init__(self, data: pd.DataFrame):
        """
        Embedding.

        :param data: DataFrame to use as term to vector mapping.
        """
        self._data = data

    @property
    def input_dim(self) -> int:
        """:return Embedding input dimension."""
        return self._data.shape[0]

    @property
    def output_dim(self) -> int:
        """:return Embedding output dimension."""
        return self._data.shape[1]

    def build_matrix(
        self,
        term_index: typing.Union[
            dict, mz.preprocessors.units.Vocabulary.TermIndex],
        initializer=lambda: np.random.uniform(-0.2, 0.2)
    ) -> np.ndarray:
        """
        Build a matrix using `term_index`.

        :param term_index: A `dict` or `TermIndex` to build with.
        :param initializer: A callable that returns a default value for missing
            terms in data. (default: a random uniform distribution in range)
            `(-0.2, 0.2)`).
        :return: A matrix.
        """
        input_dim = len(term_index) + 1

        matrix = np.empty((input_dim, self.output_dim))
        for index in np.ndindex(*matrix.shape):
            matrix[index] = initializer()

        valid_keys = set(self._data.index)
        for term, index in term_index.items():
            if term in valid_keys:
                matrix[index] = self._data.loc[term]

        return matrix


def load_from_file(file_path: str, mode: str = 'word2vec') -> Embedding:
    """
    Load embedding from `file_path`.

    :param file_path: Path to file.
    :param mode: Embedding file format mode, one of 'word2vec' or 'glove'.
        (default: 'word2vec')
    :return: An :class:`matchzoo.embedding.Embedding` instance.
    """
    if mode == 'word2vec':
        data = pd.read_csv(file_path,
                           sep=" ",
                           index_col=0,
                           header=None,
                           skiprows=1)
    elif mode == 'glove':
        data = pd.read_csv(file_path,
                           sep=" ",
                           index_col=0,
                           header=None,
                           quoting=csv.QUOTE_NONE)
    else:
        raise TypeError(f"{mode} is not a supported embedding type."
                        f"`word2vec` or `glove` expected.")
    return Embedding(data)



from .stateful_unit import StatefulUnit


class Vocabulary(StatefulUnit):
    """
    Vocabulary class.

    Examples:
        >>> vocab = Vocabulary()
        >>> vocab.fit(['A', 'B', 'C', 'D', 'E'])
        >>> term_index = vocab.state['term_index']
        >>> term_index  # doctest: +SKIP
        {'E': 1, 'C': 2, 'D': 3, 'A': 4, 'B': 5}
        >>> index_term = vocab.state['index_term']
        >>> index_term  # doctest: +SKIP
        {1: 'C', 2: 'A', 3: 'E', 4: 'B', 5: 'D'}

        >>> term_index['out-of-vocabulary-term']
        0
        >>> index_term[0]
        ''
        >>> index_term[42]
        Traceback (most recent call last):
            ...
        KeyError: 42
        >>> a_index = term_index['A']
        >>> c_index = term_index['C']
        >>> vocab.transform(['C', 'A', 'C']) == [c_index, a_index, c_index]
        True
        >>> vocab.transform(['C', 'A', 'OOV']) == [c_index, a_index, 0]
        True
        >>> indices = vocab.transform(list('ABCDDZZZ'))
        >>> ''.join(vocab.state['index_term'][i] for i in indices)
        'ABCDD'

    """

    class IndexTerm(dict):
        """Map index to term."""

        def __missing__(self, key):
            """Map out-of-vocabulary indices to empty string."""
            if key == 0:
                return ''
            else:
                raise KeyError(key)

    class TermIndex(dict):
        """Map term to index."""

        def __missing__(self, key):
            """Map out-of-vocabulary terms to index 0."""
            return 0

    def fit(self, tokens: list):
        """Build a :class:`TermIndex` and a :class:`IndexTerm`."""
        self._state['term_index'] = self.TermIndex()
        self._state['index_term'] = self.IndexTerm()
        terms = set(tokens)
        for index, term in enumerate(terms):
            self._state['term_index'][term] = index + 1
            self._state['index_term'][index + 1] = term

    def transform(self, input_: list) -> list:
        """Transform a list of tokens to corresponding indices."""
        return [self._state['term_index'][token] for token in input_]



"""Ranking task."""

from matchzoo.engine import base_task


class Ranking(base_task.BaseTask):
    """Ranking Task.

    Examples:
        >>> ranking_task = Ranking()
        >>> ranking_task.metrics = ['map', 'ndcg']
        >>> ranking_task.output_shape
        (1,)
        >>> ranking_task.output_dtype
        <class 'float'>
        >>> print(ranking_task)
        Ranking Task

    """

    @classmethod
    def list_available_losses(cls) -> list:
        """:return: a list of available losses."""
        return ['mse']

    @classmethod
    def list_available_metrics(cls) -> list:
        """:return: a list of available metrics."""
        return ['map']

    @property
    def output_shape(self) -> tuple:
        """:return: output shape of a single sample of the task."""
        return 1,

    @property
    def output_dtype(self):
        """:return: target data type, expect `float` as output."""
        return float

    def __str__(self):
        """:return: Task name as string."""
        return 'Ranking Task'



"""An implementation of CDSSM (CLSM) model."""
import typing

import keras
from keras.models import Model

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo import preprocessors
from matchzoo.utils import TensorType


class CDSSM(BaseModel):
    """
    CDSSM Model implementation.

    Learning Semantic Representations Using Convolutional Neural Networks
    for Web Search. (2014a)
    A Latent Semantic Model with Convolutional-Pooling Structure for
    Information Retrieval. (2014b)

    Examples:
        >>> model = CDSSM()
        >>> model.params['optimizer'] = 'adam'
        >>> model.params['filters'] =  32
        >>> model.params['kernel_size'] = 3
        >>> model.params['conv_activation_func'] = 'relu'
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        # set :attr:`with_multi_layer_perceptron` to False to support
        # user-defined variable dense layer units
        params = super().get_default_params(with_multi_layer_perceptron=True)
        params.add(Param(name='filters', value=32,
                         desc="Number of filters in the 1D convolution "
                              "layer."))
        params.add(Param(name='kernel_size', value=3,
                         desc="Number of kernel size in the 1D "
                              "convolution layer."))
        params.add(Param(name='strides', value=1,
                         desc="Strides in the 1D convolution layer."))
        params.add(Param(name='padding', value='same',
                         desc="The padding mode in the convolution "
                              "layer. It should be one of `same`, "
                              "`valid`, ""and `causal`."))
        params.add(Param(name='conv_activation_func', value='relu',
                         desc="Activation function in the convolution"
                              " layer."))
        params.add(Param(name='w_initializer', value='glorot_normal'))
        params.add(Param(name='b_initializer', value='zeros'))
        params.add(Param(name='dropout_rate', value=0.3,
                         desc="The dropout rate."))
        return params

    def _create_base_network(self) -> typing.Callable:
        """
        Apply conv and maxpooling operation towards to each letter-ngram.

        The input shape is `fixed_text_length`*`number of letter-ngram`,
        as described in the paper, `n` is 3, `number of letter-trigram`
        is about 30,000 according to their observation.

        :return: Wrapped Keras `Layer` as CDSSM network, tensor in tensor out.
        """

        def _wrapper(x: TensorType):
            # Apply 1d convolutional on each word_ngram (lt).
            # Input shape: (batch_size, num_tri_letters, 90000)
            # Sequence of num_tri_letters vectors of 90000d vectors.
            x = keras.layers.Conv1D(
                filters=self._params['filters'],
                kernel_size=self._params['kernel_size'],
                strides=self._params['strides'],
                padding=self._params['padding'],
                activation=self._params['conv_activation_func'],
                kernel_initializer=self._params['w_initializer'],
                bias_initializer=self._params['b_initializer'])(x)
            # Apply max pooling by take max at each dimension across
            # all word_trigram features.
            x = keras.layers.Dropout(self._params['dropout_rate'])(x)
            x = keras.layers.GlobalMaxPool1D()(x)
            # Apply a none-linear transformation use a tanh layer.
            x = self._make_multi_layer_perceptron_layer()(x)
            return x

        return _wrapper

    def build(self):
        """
        Build model structure.

        CDSSM use Siamese architecture.
        """
        base_network = self._create_base_network()
        # Left input and right input.
        input_left, input_right = self._make_inputs()
        # Process left & right input.
        x = [base_network(input_left),
             base_network(input_right)]
        # Dot product with cosine similarity.
        x = keras.layers.Dot(axes=[1, 1], normalize=True)(x)
        x_out = self._make_output_layer()(x)
        self._backend = Model(inputs=[input_left, input_right],
                              outputs=x_out)

    @classmethod
    def get_default_preprocessor(cls):
        """:return: Default preprocessor."""
        return preprocessors.CDSSMPreprocessor()

    def guess_and_fill_missing_params(self, verbose: int = 1):
        """
        Guess and fill missing parameters in :attr:`params`.

        Use this method to automatically fill-in hyper parameters.
        This involves some guessing so the parameter it fills could be
        wrong. For example, the default task is `Ranking`, and if we do not
        set it to `Classification` manually for data packs prepared for
        classification, then the shape of the model output and the data will
        mismatch.

        :param verbose: Verbosity.
        """
        self._params.get('input_shapes').set_default([(10, 30),
                                                      (10, 30)], verbose)
        super().guess_and_fill_missing_params(verbose)



import abc
import typing


class Unit(metaclass=abc.ABCMeta):
    """Process unit do not persive state (i.e. do not need fit)."""

    @abc.abstractmethod
    def transform(self, input_: typing.Any):
        """Abstract base method, need to be implemented in subclass."""



import pytest

import matchzoo as mz


@pytest.fixture
def term_index():
    return {'G': 1, 'C': 2, 'D': 3, 'A': 4, '[PAD]': 0}


def test_embedding(term_index):
    embed = mz.embedding.load_from_file(mz.datasets.embeddings.EMBED_RANK)
    matrix = embed.build_matrix(term_index)
    assert matrix.shape == (len(term_index) + 1, 50)
    embed = mz.embedding.load_from_file(mz.datasets.embeddings.EMBED_10_GLOVE,
                                        mode='glove')
    matrix = embed.build_matrix(term_index)
    assert matrix.shape == (len(term_index) + 1, 10)
    assert embed.input_dim == 5



import typing

import numpy as np

import matchzoo as mz
from matchzoo.engine.base_task import BaseTask
from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.base_preprocessor import BasePreprocessor
from matchzoo.data_generator import DataGeneratorBuilder


class Preparer(object):
    """
    Unified setup processes of all MatchZoo models.

    `config` is used to control specific behaviors. The default `config`
    will be updated accordingly if a `config` dictionary is passed. e.g. to
    override the default `bin_size`, pass `config={'bin_size': 15}`.

    See `tutorials/automation.ipynb` for a detailed walkthrough on usage.

    Default `config`:

    {
        # pair generator builder kwargs
        'num_dup': 1,

        # histogram unit of DRMM
        'bin_size': 30,
        'hist_mode': 'LCH',

        # dynamic Pooling of MatchPyramid
        'compress_ratio_left': 1.0,
        'compress_ratio_right': 1.0,

        # if no `matchzoo.Embedding` is passed to `tune`
        'embedding_output_dim': 50
    }

    :param task: Task.
    :param config: Configuration of specific behaviors.

    Example:
        >>> import matchzoo as mz
        >>> task = mz.tasks.Ranking(loss=mz.losses.RankCrossEntropyLoss())
        >>> preparer = mz.auto.Preparer(task)
        >>> model_class = mz.models.DenseBaseline
        >>> train_raw = mz.datasets.toy.load_data('train', 'ranking')
        >>> model, prpr, gen_builder, matrix = preparer.prepare(model_class,
        ...                                                     train_raw)
        >>> model.params.completed()
        True

    """

    def __init__(
        self,
        task: BaseTask,
        config: typing.Optional[dict] = None
    ):
        """Init."""
        self._task = task
        self._config = self.get_default_config()
        if config:
            self._config.update(config)

        self._infer_num_neg()

    def prepare(
        self,
        model_class: typing.Type[BaseModel],
        data_pack: mz.DataPack,
        preprocessor: typing.Optional[BasePreprocessor] = None,
        embedding: typing.Optional['mz.Embedding'] = None,
    ) -> typing.Tuple[
        BaseModel,
        BasePreprocessor,
        DataGeneratorBuilder,
        np.ndarray
    ]:
        """
        Prepare.

        :param model_class: Model class.
        :param data_pack: DataPack used to fit the preprocessor.
        :param preprocessor: Preprocessor used to fit the `data_pack`.
            (default: the default preprocessor of `model_class`)
        :param embedding: Embedding to build a embedding matrix. If not set,
            then a correctly shaped randomized matrix will be built.

        :return: A tuple of `(model, preprocessor, data_generator_builder,
            embedding_matrix)`.

        """
        if not preprocessor:
            preprocessor = model_class.get_default_preprocessor()

        if issubclass(model_class, (mz.models.DSSM, mz.models.CDSSM)):
            preprocessor.with_word_hashing = False

        preprocessor.fit(data_pack, verbose=0)

        model, embedding_matrix = self._build_model(
            model_class,
            preprocessor,
            embedding
        )

        data_gen_builder = self._build_data_gen_builder(
            model,
            embedding_matrix,
            preprocessor
        )

        return (
            model,
            preprocessor,
            data_gen_builder,
            embedding_matrix
        )

    def _build_model(
        self,
        model_class,
        preprocessor,
        embedding
    ) -> typing.Tuple[BaseModel, np.ndarray]:

        model = model_class()
        model.params['task'] = self._task
        model.params.update(preprocessor.context)

        if 'with_embedding' in model.params:
            embedding_matrix = self._build_matrix(preprocessor, embedding)
            model.params['embedding_input_dim'] = embedding_matrix.shape[0]
            model.params['embedding_output_dim'] = embedding_matrix.shape[1]
        else:
            embedding_matrix = None

        self._handle_drmm_input_shapes(model)

        assert model.params.completed()
        model.build()
        model.compile()

        if 'with_embedding' in model.params:
            model.load_embedding_matrix(embedding_matrix)

        return model, embedding_matrix

    def _handle_drmm_input_shapes(self, model):
        if isinstance(model, mz.models.DRMM):
            left = model.params['input_shapes'][0]
            right = left + (self._config['bin_size'],)
            model.params['input_shapes'] = (left, right)

    def _build_matrix(self, preprocessor, embedding):
        if embedding:
            vocab_unit = preprocessor.context['vocab_unit']
            term_index = vocab_unit.state['term_index']
            return embedding.build_matrix(term_index)
        else:
            matrix_shape = (
                preprocessor.context['vocab_size'],
                self._config['embedding_output_dim']
            )
            return np.random.uniform(-0.2, 0.2, matrix_shape)

    def _build_data_gen_builder(self, model, embedding_matrix, preprocessor):
        builder_kwargs = dict(callbacks=[])

        if isinstance(self._task.loss, (mz.losses.RankHingeLoss,
                                        mz.losses.RankCrossEntropyLoss)):
            builder_kwargs.update(dict(
                mode='pair',
                num_dup=self._config['num_dup'],
                num_neg=self._config['num_neg']
            ))

        if isinstance(model, mz.models.DRMM):
            histo_callback = mz.data_generator.callbacks.Histogram(
                embedding_matrix=embedding_matrix,
                bin_size=self._config['bin_size'],
                hist_mode=self._config['hist_mode']
            )
            builder_kwargs['callbacks'].append(histo_callback)

        if isinstance(model, mz.models.MatchPyramid):
            dpool_callback = mz.data_generator.callbacks.DynamicPooling(
                fixed_length_left=model.params['input_shapes'][0][0],
                fixed_length_right=model.params['input_shapes'][1][0],
                compress_ratio_left=self._config['compress_ratio_left'],
                compress_ratio_right=self._config['compress_ratio_right']
            )
            builder_kwargs['callbacks'].append(dpool_callback)

        if isinstance(model, (mz.models.DSSM, mz.models.CDSSM)):
            term_index = preprocessor.context['vocab_unit'].state['term_index']
            hashing_unit = mz.preprocessors.units.WordHashing(term_index)
            hashing_callback = mz.data_generator.callbacks.LambdaCallback(
                on_batch_data_pack=lambda data_pack:
                data_pack.apply_on_text(
                    func=hashing_unit.transform,
                    inplace=True,
                    verbose=0
                )
            )
            builder_kwargs['callbacks'].append(hashing_callback)

        return DataGeneratorBuilder(**builder_kwargs)

    def _infer_num_neg(self):
        if isinstance(self._task.loss, (mz.losses.RankHingeLoss,
                                        mz.losses.RankCrossEntropyLoss)):
            self._config['num_neg'] = self._task.loss.num_neg

    @classmethod
    def get_default_config(cls) -> dict:
        """Default config getter."""
        return {
            # pair generator builder kwargs
            'num_dup': 1,

            # histogram unit of DRMM
            'bin_size': 30,
            'hist_mode': 'LCH',

            # dynamic Pooling of MatchPyramid
            'compress_ratio_left': 1.0,
            'compress_ratio_right': 1.0,

            # if no `matchzoo.Embedding` is passed to `tune`
            'embedding_output_dim': 100
        }



"""Naive model with a simplest structure for testing purposes."""

import keras

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine import hyper_spaces


class Naive(BaseModel):
    """
    Naive model with a simplest structure for testing purposes.

    Bare minimum functioning model. The best choice to get things rolling.
    The worst choice to fit and evaluate performance.
    """

    @classmethod
    def get_default_params(cls):
        """Default parameters."""
        params = super().get_default_params()
        params.get('optimizer').hyper_space = \
            hyper_spaces.choice(['adam', 'adagrad', 'rmsprop'])
        return params

    def build(self):
        """Build."""
        x_in = self._make_inputs()
        x = keras.layers.concatenate(x_in)
        x_out = self._make_output_layer()(x)
        self._backend = keras.models.Model(inputs=x_in, outputs=x_out)



from matchzoo.data_generator.callbacks.callback import Callback


class LambdaCallback(Callback):
    """
    LambdaCallback. Just a shorthand for creating a callback class.

    See :class:`matchzoo.data_generator.callbacks.Callback` for more details.

    Example:
        >>> from matchzoo.data_generator.callbacks import LambdaCallback
        >>> callback = LambdaCallback(on_batch_unpacked=print)
        >>> callback.on_batch_unpacked('x', 'y')
        x y

    """

    def __init__(self, on_batch_data_pack=None, on_batch_unpacked=None):
        """Init."""
        self._on_batch_unpacked = on_batch_unpacked
        self._on_batch_data_pack = on_batch_data_pack

    def on_batch_data_pack(self, data_pack):
        """`on_batch_data_pack`."""
        if self._on_batch_data_pack:
            self._on_batch_data_pack(data_pack)

    def on_batch_unpacked(self, x, y):
        """`on_batch_unpacked`."""
        if self._on_batch_unpacked:
            self._on_batch_unpacked(x, y)



import typing

import matchzoo as mz
from .preparer import Preparer
from matchzoo.engine.base_task import BaseTask
from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.base_preprocessor import BasePreprocessor


def prepare(
    task: BaseTask,
    model_class: typing.Type[BaseModel],
    data_pack: mz.DataPack,
    preprocessor: typing.Optional[BasePreprocessor] = None,
    embedding: typing.Optional['mz.Embedding'] = None,
    config: typing.Optional[dict] = None,
):
    """
    A simple shorthand for using :class:`matchzoo.Preparer`.

    `config` is used to control specific behaviors. The default `config`
    will be updated accordingly if a `config` dictionary is passed. e.g. to
    override the default `bin_size`, pass `config={'bin_size': 15}`.

    :param task: Task.
    :param model_class: Model class.
    :param data_pack: DataPack used to fit the preprocessor.
    :param preprocessor: Preprocessor used to fit the `data_pack`.
        (default: the default preprocessor of `model_class`)
    :param embedding: Embedding to build a embedding matrix. If not set,
        then a correctly shaped randomized matrix will be built.
    :param config: Configuration of specific behaviors. (default: return
        value of `mz.Preparer.get_default_config()`)

    :return: A tuple of `(model, preprocessor, data_generator_builder,
        embedding_matrix)`.

    """
    preparer = Preparer(task=task, config=config)
    return preparer.prepare(
        model_class=model_class,
        data_pack=data_pack,
        preprocessor=preprocessor,
        embedding=embedding
    )



"""DUET Model."""

import keras
import keras.backend as K
import tensorflow as tf

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class DUET(BaseModel):
    """
    DUET Model.

    Examples:
        >>> model = DUET()
        >>> model.params['embedding_input_dim'] = 1000
        >>> model.params['embedding_output_dim'] = 300
        >>> model.params['lm_filters'] = 32
        >>> model.params['lm_hidden_sizes'] = [64, 32]
        >>> model.params['dropout_rate'] = 0.5
        >>> model.params['dm_filters'] = 32
        >>> model.params['dm_kernel_size'] = 3
        >>> model.params['dm_d_mpool'] = 4
        >>> model.params['dm_hidden_sizes'] = [64, 32]
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls):
        """Get default parameters."""
        params = super().get_default_params(with_embedding=True)
        params.add(Param(name='lm_filters', value=32,
                         desc="Filter size of 1D convolution layer in "
                              "the local model."))
        params.add(Param(name='lm_hidden_sizes', value=[32],
                         desc="A list of hidden size of the MLP layer "
                              "in the local model."))
        params.add(Param(name='dm_filters', value=32,
                         desc="Filter size of 1D convolution layer in "
                              "the distributed model."))
        params.add(Param(name='dm_kernel_size', value=3,
                         desc="Kernel size of 1D convolution layer in "
                              "the distributed model."))
        params.add(Param(name='dm_q_hidden_size', value=32,
                         desc="Hidden size of the MLP layer for the "
                              "left text in the distributed model."))
        params.add(Param(name='dm_d_mpool', value=3,
                         desc="Max pooling size for the right text in "
                              "the distributed model."))
        params.add(Param(name='dm_hidden_sizes', value=[32],
                         desc="A list of hidden size of the MLP layer "
                              "in the distributed model."))
        params.add(Param(name='padding', value='same',
                         desc="The padding mode in the convolution "
                              "layer. It should be one of `same`, "
                              "`valid`, ""and `causal`."))
        params.add(Param(name='activation_func', value='relu',
                         desc="Activation function in the convolution"
                              " layer."))
        params.add(Param(
            name='dropout_rate', value=0.5,
            hyper_space=hyper_spaces.quniform(low=0.0, high=0.8,
                                              q=0.02),
            desc="The dropout rate."))
        return params

    def build(self):
        """Build model."""
        query, doc = self._make_inputs()

        embedding = self._make_embedding_layer()
        q_embed = embedding(query)
        d_embed = embedding(doc)

        lm_xor = keras.layers.Lambda(self._xor_match)([query, doc])
        lm_conv = keras.layers.Conv1D(
            self._params['lm_filters'],
            self._params['input_shapes'][1][0],
            padding=self._params['padding'],
            activation=self._params['activation_func']
        )(lm_xor)

        lm_conv = keras.layers.Dropout(self._params['dropout_rate'])(
            lm_conv)
        lm_feat = keras.layers.Reshape((-1,))(lm_conv)
        for hidden_size in self._params['lm_hidden_sizes']:
            lm_feat = keras.layers.Dense(
                hidden_size,
                activation=self._params['activation_func']
            )(lm_feat)
        lm_drop = keras.layers.Dropout(self._params['dropout_rate'])(
            lm_feat)
        lm_score = keras.layers.Dense(1)(lm_drop)

        dm_q_conv = keras.layers.Conv1D(
            self._params['dm_filters'],
            self._params['dm_kernel_size'],
            padding=self._params['padding'],
            activation=self._params['activation_func']
        )(q_embed)
        dm_q_conv = keras.layers.Dropout(self._params['dropout_rate'])(
            dm_q_conv)
        dm_q_mp = keras.layers.MaxPooling1D(
            pool_size=self._params['input_shapes'][0][0])(dm_q_conv)
        dm_q_rep = keras.layers.Reshape((-1,))(dm_q_mp)
        dm_q_rep = keras.layers.Dense(self._params['dm_q_hidden_size'])(
            dm_q_rep)
        dm_q_rep = keras.layers.Lambda(lambda x: tf.expand_dims(x, 1))(
            dm_q_rep)

        dm_d_conv1 = keras.layers.Conv1D(
            self._params['dm_filters'],
            self._params['dm_kernel_size'],
            padding=self._params['padding'],
            activation=self._params['activation_func']
        )(d_embed)
        dm_d_conv1 = keras.layers.Dropout(self._params['dropout_rate'])(
            dm_d_conv1)
        dm_d_mp = keras.layers.MaxPooling1D(
            pool_size=self._params['dm_d_mpool'])(dm_d_conv1)
        dm_d_conv2 = keras.layers.Conv1D(
            self._params['dm_filters'], 1,
            padding=self._params['padding'],
            activation=self._params['activation_func']
        )(dm_d_mp)
        dm_d_conv2 = keras.layers.Dropout(self._params['dropout_rate'])(
            dm_d_conv2)

        h_dot = keras.layers.Lambda(self._hadamard_dot)([dm_q_rep, dm_d_conv2])
        dm_feat = keras.layers.Reshape((-1,))(h_dot)
        for hidden_size in self._params['dm_hidden_sizes']:
            dm_feat = keras.layers.Dense(hidden_size)(dm_feat)
        dm_feat_drop = keras.layers.Dropout(self._params['dropout_rate'])(
            dm_feat)
        dm_score = keras.layers.Dense(1)(dm_feat_drop)

        add = keras.layers.Add()([lm_score, dm_score])
        out = self._make_output_layer()(add)
        self._backend = keras.Model(inputs=[query, doc], outputs=out)

    @classmethod
    def _xor_match(cls, x):
        t1 = x[0]
        t2 = x[1]
        t1_shape = t1.get_shape()
        t2_shape = t2.get_shape()
        t1_expand = K.tf.stack([t1] * t2_shape[1], 2)
        t2_expand = K.tf.stack([t2] * t1_shape[1], 1)
        out_bool = K.tf.equal(t1_expand, t2_expand)
        out = K.tf.cast(out_bool, K.tf.float32)
        return out

    @classmethod
    def _hadamard_dot(cls, x):
        x1 = x[0]
        x2 = x[1]
        out = x1 * x2
        return out



import numpy as np

from matchzoo.engine.base_metric import sort_and_couple
from matchzoo import metrics


def test_sort_and_couple():
    l = [0, 1, 2]
    s = [0.1, 0.4, 0.2]
    c = sort_and_couple(l, s)
    assert (c == np.array([(1, 0.4), (2, 0.2), (0, 0.1)])).all()


def test_mean_reciprocal_rank():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    assert metrics.MeanReciprocalRank()(label, score) == 1


def test_precision_at_k():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    assert metrics.Precision(k=1)(label, score) == 1.
    assert metrics.Precision(k=2)(label, score) == 1.
    assert round(metrics.Precision(k=3)(label, score), 2) == 0.67


def test_average_precision():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    assert round(metrics.AveragePrecision()(label, score), 2) == 0.89


def test_mean_average_precision():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    assert metrics.MeanAveragePrecision()(label, score) == 1.


def test_dcg_at_k():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    dcg = metrics.DiscountedCumulativeGain
    assert round(dcg(k=1)(label, score), 2) == 1.44
    assert round(dcg(k=2)(label, score), 2) == 4.17
    assert round(dcg(k=3)(label, score), 2) == 4.17


def test_ndcg_at_k():
    label = [0, 1, 2]
    score = [0.1, 0.4, 0.2]
    ndcg = metrics.NormalizedDiscountedCumulativeGain
    assert round(ndcg(k=1)(label, score), 2) == 0.33
    assert round(ndcg(k=2)(label, score), 2) == 0.80
    assert round(ndcg(k=3)(label, score), 2) == 0.80



"""CDSSM Preprocessor."""

from tqdm import tqdm

from . import units
from .chain_transform import chain_transform
from matchzoo import DataPack
from matchzoo.engine.base_preprocessor import BasePreprocessor
from .build_vocab_unit import build_vocab_unit

tqdm.pandas()


class CDSSMPreprocessor(BasePreprocessor):
    """CDSSM Model preprocessor."""

    def __init__(self,
                 fixed_length_left: int = 10,
                 fixed_length_right: int = 40,
                 with_word_hashing: bool = True):
        """
        CDSSM Model preprocessor.

        The word hashing step could eats up a lot of memory. To workaround
        this problem, set `with_word_hashing` to `False` and use a
        :class:`matchzoo.DynamicDataGenerator` with a
        :class:`matchzoo.preprocessor.units.WordHashing`.

        TODO: doc here.

        :param with_word_hashing: Include a word hashing step if `True`.

        Example:
            >>> import matchzoo as mz
            >>> train_data = mz.datasets.toy.load_data()
            >>> test_data = mz.datasets.toy.load_data(stage='test')
            >>> cdssm_preprocessor = mz.preprocessors.CDSSMPreprocessor()
            >>> train_data_processed = cdssm_preprocessor.fit_transform(
            ...     train_data, verbose=0
            ... )
            >>> type(train_data_processed)
            <class 'matchzoo.data_pack.data_pack.DataPack'>
            >>> test_data_transformed = cdssm_preprocessor.transform(test_data,
            ...                                                      verbose=0)
            >>> type(test_data_transformed)
            <class 'matchzoo.data_pack.data_pack.DataPack'>

        """
        super().__init__()
        self._fixed_length_left = fixed_length_left
        self._fixed_length_right = fixed_length_right
        self._left_fixedlength_unit = units.FixedLength(
            self._fixed_length_left,
            pad_value='0', pad_mode='post'
        )
        self._right_fixedlength_unit = units.FixedLength(
            self._fixed_length_right,
            pad_value='0', pad_mode='post'
        )
        self._with_word_hashing = with_word_hashing

    def fit(self, data_pack: DataPack, verbose: int = 1):
        """
        Fit pre-processing context for transformation.

        :param verbose: Verbosity.
        :param data_pack: Data_pack to be preprocessed.
        :return: class:`CDSSMPreprocessor` instance.
        """
        fit_units = self._default_units() + [units.NgramLetter()]
        func = chain_transform(fit_units)
        data_pack = data_pack.apply_on_text(func, verbose=verbose)
        vocab_unit = build_vocab_unit(data_pack, verbose=verbose)

        self._context['vocab_unit'] = vocab_unit
        vocab_size = len(vocab_unit.state['term_index']) + 1
        self._context['input_shapes'] = [
            (self._fixed_length_left, vocab_size),
            (self._fixed_length_right, vocab_size)
        ]
        return self

    def transform(self, data_pack: DataPack, verbose: int = 1) -> DataPack:
        """
        Apply transformation on data, create `letter-ngram` representation.

        :param data_pack: Inputs to be preprocessed.
        :param verbose: Verbosity.

        :return: Transformed data as :class:`DataPack` object.
        """
        data_pack = data_pack.copy()
        func = chain_transform(self._default_units())
        data_pack.apply_on_text(func, inplace=True, verbose=verbose)
        data_pack.apply_on_text(self._left_fixedlength_unit.transform,
                                mode='left', inplace=True, verbose=verbose)
        data_pack.apply_on_text(self._right_fixedlength_unit.transform,
                                mode='right', inplace=True, verbose=verbose)
        post_units = [units.NgramLetter(reduce_dim=False)]
        if self._with_word_hashing:
            term_index = self._context['vocab_unit'].state['term_index']
            post_units.append(units.WordHashing(term_index))
        data_pack.apply_on_text(chain_transform(post_units),
                                inplace=True, verbose=verbose)
        return data_pack

    @classmethod
    def _default_units(cls) -> list:
        """Prepare needed process units."""
        return [
            units.Tokenize(),
            units.Lowercase(),
            units.PuncRemoval(),
            units.StopRemoval(),
        ]

    @property
    def with_word_hashing(self):
        """`with_word_hashing` getter."""
        return self._with_word_hashing

    @with_word_hashing.setter
    def with_word_hashing(self, value):
        """`with_word_hashing` setter."""
        self._with_word_hashing = value



import nltk

from .unit import Unit


class Lemmatization(Unit):
    """Process unit for token lemmatization."""

    def transform(self, input_: list) -> list:
        """
        Lemmatization a sequence of tokens.

        :param input_: list of tokens to be lemmatized.

        :return tokens: list of lemmatizd tokens.
        """
        lemmatizer = nltk.WordNetLemmatizer()
        return [lemmatizer.lemmatize(token, pos='v') for token in input_]



from matchzoo.engine.base_model import BaseModel
from .callback import Callback


class LambdaCallback(Callback):
    """
    LambdaCallback. Just a shorthand for creating a callback class.

    See :class:`matchzoo.tuner.callbacks.Callback` for more details.
    """

    def __init__(
        self,
        on_run_start=None,
        on_build_end=None,
        on_result_end=None
    ):
        """Init."""
        self._on_run_start = on_run_start
        self._on_build_end = on_build_end
        self._on_result_end = on_result_end

    def on_run_start(self, tuner, sample: dict):
        """`on_run_start`."""
        if self._on_run_start:
            self._on_run_start(tuner, sample)

    def on_build_end(self, tuner, model: BaseModel):
        """`on_build_end`."""
        if self._on_build_end:
            self._on_build_end(tuner, model)

    def on_run_end(self, tuner, model: BaseModel, result: dict):
        """`on_run_end`."""
        if self._on_result_end:
            self._on_result_end(tuner, model, result)



import pytest

import matchzoo as mz


@pytest.fixture(scope='module')
def tuner():
    model = mz.models.DenseBaseline()
    prpr = model.get_default_preprocessor()
    train_raw = mz.datasets.toy.load_data('train')
    dev_raw = mz.datasets.toy.load_data('dev')
    prpr.fit(train_raw)
    model.params.update(prpr.context)
    model.guess_and_fill_missing_params()
    return mz.auto.Tuner(
        params=model.params,
        train_data=prpr.transform(train_raw, verbose=0),
        test_data=prpr.transform(dev_raw, verbose=0)
    )


@pytest.mark.parametrize('attr', [
    'params',
    'train_data',
    'test_data',
    'fit_kwargs',
    'evaluate_kwargs',
    'metric',
    'mode',
    'num_runs',
    'callbacks',
    'verbose'
])
def test_getters_setters(tuner, attr):
    val = getattr(tuner, attr)
    setattr(tuner, attr, val)
    assert getattr(tuner, attr) is val


def test_tuning(tuner):
    tuner.num_runs = 1
    assert tuner.tune()



"""Callbacks."""
import typing
from pathlib import Path

import numpy as np
import keras

import matchzoo
from matchzoo.engine.base_model import BaseModel


class EvaluateAllMetrics(keras.callbacks.Callback):
    """
    Callback to evaluate all metrics.

    MatchZoo metrics can not be evaluated batch-wise since they require
    dataset-level information. As a result, MatchZoo metrics are not
    evaluated automatically when a Model `fit`. When this callback is used,
    all metrics, including MatchZoo metrics and Keras metrics, are evluated
    once every `once_every` epochs.

    :param model: Model to evaluate.
    :param x: X.
    :param y: y.
    :param once_every: Evaluation only triggers when `epoch % once_every == 0`.
        (default: 1, i.e. evaluate on every epoch's end)
    :param batch_size: Number of samples per evaluation. This only affects the
        evaluation of Keras metrics, since MatchZoo metrics are always
        evaluated using the full data.
    :param model_save_path: Directory path to save the model after each
        evaluate callback, (default: None, i.e., no saving.)
    :param verbose: Verbosity.
    """

    def __init__(
        self,
        model: 'BaseModel',
        x: typing.Union[np.ndarray, typing.List[np.ndarray]],
        y: np.ndarray,
        once_every: int = 1,
        batch_size: int = 128,
        model_save_path: str = None,
        verbose=1
    ):
        """Initializer."""
        super().__init__()
        self._model = model
        self._dev_x = x
        self._dev_y = y
        self._valid_steps = once_every
        self._batch_size = batch_size
        self._model_save_path = model_save_path
        self._verbose = verbose

    def on_epoch_end(self, epoch: int, logs: dict = None):
        """
        Called at the end of en epoch.

        :param epoch: integer, index of epoch.
        :param logs: dictionary of logs.
        :return: dictionary of logs.
        """
        if (epoch + 1) % self._valid_steps == 0:
            val_logs = self._model.evaluate(self._dev_x, self._dev_y,
                                            self._batch_size)
            if self._verbose:
                print('Validation: ' + ' - '.join(
                    f'{k}: {v}' for k, v in val_logs.items()))
            for k, v in val_logs.items():
                logs[k] = v
            if self._model_save_path:
                curr_path = self._model_save_path + str('%d/' % (epoch + 1))
                self._model.save(curr_path)



"""Matchzoo DataPack, pair-wise tuple (feature) and context as input."""

import typing
import inspect
from pathlib import Path
import functools

import dill
from tqdm import tqdm
import numpy as np
import pandas as pd

import matchzoo

tqdm.pandas()


def _convert_to_list_index(
    index: typing.Union[int, slice, np.array],
    length: int
):
    if isinstance(index, int):
        index = [index]
    elif isinstance(index, slice):
        index = list(range(*index.indices(length)))
    return index


class DataPack(object):
    """
    Matchzoo :class:`DataPack` data structure, store dataframe and context.

    `DataPack` is a MatchZoo native data structure that most MatchZoo data
    handling processes build upon. A `DataPack` consists of three parts:
    `left`, `right` and `relation`, each one of is a `pandas.DataFrame`.

    :param relation: Store the relation between left document
        and right document use ids.
    :param left: Store the content or features for id_left.
    :param right: Store the content or features for
        id_right.

    Example:
        >>> left = [
        ...     ['qid1', 'query 1'],
        ...     ['qid2', 'query 2']
        ... ]
        >>> right = [
        ...     ['did1', 'document 1'],
        ...     ['did2', 'document 2']
        ... ]
        >>> relation = [['qid1', 'did1', 1], ['qid2', 'did2', 1]]
        >>> relation_df = pd.DataFrame(relation)
        >>> left = pd.DataFrame(left)
        >>> right = pd.DataFrame(right)
        >>> dp = DataPack(
        ...     relation=relation_df,
        ...     left=left,
        ...     right=right,
        ... )
        >>> len(dp)
        2
    """

    DATA_FILENAME = 'data.dill'

    def __init__(
        self,
        relation: pd.DataFrame,
        left: pd.DataFrame,
        right: pd.DataFrame
    ):
        """:class:`DataPack` initializer."""
        self._relation = relation
        self._left = left
        self._right = right

    @property
    def has_label(self) -> bool:
        """:return: `True` if `label` column exists, `False` other wise."""
        return 'label' in self._relation.columns

    def __len__(self) -> int:
        """Get numer of rows in the class:`DataPack` object."""
        return self._relation.shape[0]

    @property
    def frame(self) -> 'DataPack.FrameView':
        """
        View the data pack as a :class:`pandas.DataFrame`.

        Returned data frame is created by merging the left data frame,
        the right dataframe and the relation data frame. Use `[]` to access
        an item or a slice of items.

        :return: A :class:`matchzoo.DataPack.FrameView` instance.

        Example:
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> type(data_pack.frame)
            <class 'matchzoo.data_pack.data_pack.DataPack.FrameView'>
            >>> frame_slice = data_pack.frame[0:5]
            >>> type(frame_slice)
            <class 'pandas.core.frame.DataFrame'>
            >>> list(frame_slice.columns)
            ['id_left', 'text_left', 'id_right', 'text_right', 'label']
            >>> full_frame = data_pack.frame()
            >>> len(full_frame) == len(data_pack)
            True

        """
        return DataPack.FrameView(self)

    def unpack(self) -> typing.Tuple[typing.Dict[str, np.array],
                                     typing.Optional[np.array]]:
        """
        Unpack the data for training.

        The return value can be directly feed to `model.fit` or
        `model.fit_generator`.

        :return: A tuple of (X, y). `y` is `None` if `self` has no label.

        Example:
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> X, y = data_pack.unpack()
            >>> type(X)
            <class 'dict'>
            >>> sorted(X.keys())
            ['id_left', 'id_right', 'text_left', 'text_right']
            >>> type(y)
            <class 'numpy.ndarray'>
            >>> X, y = data_pack.drop_label().unpack()
            >>> type(y)
            <class 'NoneType'>

        """
        frame = self.frame()

        columns = list(frame.columns)
        if self.has_label:
            columns.remove('label')
            y = np.vstack(np.asarray(frame['label']))
        else:
            y = None

        x = frame[columns].to_dict(orient='list')
        for key, val in x.items():
            x[key] = np.array(val)

        return x, y

    def __getitem__(self, index: typing.Union[int, slice, np.array]
                    ) -> 'DataPack':
        """
        Get specific item(s) as a new :class:`DataPack`.

        The returned :class:`DataPack` will be a copy of the subset of the
        original :class:`DataPack`.

        :param index: Index of the item(s) to get.
        :return: An instance of :class:`DataPack`.
        """
        index = _convert_to_list_index(index, len(self))
        relation = self._relation.loc[index].reset_index(drop=True)
        left = self._left.loc[relation['id_left'].unique()]
        right = self._right.loc[relation['id_right'].unique()]
        return DataPack(left=left.copy(),
                        right=right.copy(),
                        relation=relation.copy())

    @property
    def relation(self):
        """`relation` getter."""
        return self._relation

    @relation.setter
    def relation(self, value):
        """`relation` setter."""
        self._relation = value

    @property
    def left(self) -> pd.DataFrame:
        """Get :meth:`left` of :class:`DataPack`."""
        return self._left

    @property
    def right(self) -> pd.DataFrame:
        """Get :meth:`right` of :class:`DataPack`."""
        return self._right

    def copy(self) -> 'DataPack':
        """:return: A deep copy."""
        return DataPack(left=self._left.copy(),
                        right=self._right.copy(),
                        relation=self._relation.copy())

    def save(self, dirpath: typing.Union[str, Path]):
        """
        Save the :class:`DataPack` object.

        A saved :class:`DataPack` is represented as a directory with a
        :class:`DataPack` object (transformed user input as features and
        context), it will be saved by `pickle`.

        :param dirpath: directory path of the saved :class:`DataPack`.
        """
        dirpath = Path(dirpath)
        data_file_path = dirpath.joinpath(self.DATA_FILENAME)

        if data_file_path.exists():
            raise FileExistsError(
                f'{data_file_path} already exist, fail to save')
        elif not dirpath.exists():
            dirpath.mkdir()

        dill.dump(self, open(data_file_path, mode='wb'))

    def _optional_inplace(func):
        """
        Decorator that adds `inplace` key word argument to a method.

        Decorate any method that modifies inplace to make that inplace change
        optional.
        """
        doc = ":param inplace: `True` to modify inplace, `False` to return " \
              "a modified copy. (default: `False`)"

        def _clean(s):
            return s.replace(' ', '').replace('\n', '')

        if _clean(doc) not in _clean(inspect.getdoc(func)):
            raise NotImplementedError(
                f"`inplace` parameter of {func} not documented.\n"
                f"Please add the following line to its documentation:\n{doc}")

        @functools.wraps(func)
        def wrapper(
            self, *args, inplace: bool = False, **kwargs
        ) -> typing.Optional['DataPack']:

            if inplace:
                target = self
            else:
                target = self.copy()

            func(target, *args, **kwargs)

            if not inplace:
                return target

        return wrapper

    @_optional_inplace
    def shuffle(self):
        """
        Shuffle the data pack by shuffling the relation column.

        :param inplace: `True` to modify inplace, `False` to return a modified
            copy. (default: `False`)

        Example:
            >>> import matchzoo as mz
            >>> import numpy.random
            >>> numpy.random.seed(0)
            >>> data_pack = mz.datasets.toy.load_data()
            >>> orig_ids = data_pack.relation['id_left']
            >>> shuffled = data_pack.shuffle()
            >>> (shuffled.relation['id_left'] != orig_ids).any()
            True

        """
        self._relation = self._relation.sample(frac=1)
        self._relation.reset_index(drop=True, inplace=True)

    @_optional_inplace
    def drop_label(self):
        """
        Remove `label` column from the data pack.

        :param inplace: `True` to modify inplace, `False` to return a modified
            copy. (default: `False`)

        Example:
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> data_pack.has_label
            True
            >>> data_pack.drop_label(inplace=True)
            >>> data_pack.has_label
            False
        """
        self._relation = self._relation.drop(columns='label')

    @_optional_inplace
    def append_text_length(self, verbose=1):
        """
        Append `length_left` and `length_right` columns.

        :param inplace: `True` to modify inplace, `False` to return a modified
            copy. (default: `False`)
        :param verbose: Verbosity.

        Example:
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> 'length_left' in data_pack.frame[0].columns
            False
            >>> new_data_pack = data_pack.append_text_length(verbose=0)
            >>> 'length_left' in new_data_pack.frame[0].columns
            True
            >>> 'length_left' in data_pack.frame[0].columns
            False
            >>> data_pack.append_text_length(inplace=True, verbose=0)
            >>> 'length_left' in data_pack.frame[0].columns
            True

        """
        self.apply_on_text(len, rename=('length_left', 'length_right'),
                           inplace=True, verbose=verbose)

    @_optional_inplace
    def apply_on_text(
        self, func: typing.Callable,
        mode: str = 'both',
        rename: typing.Optional[str] = None,
        verbose: int = 1
    ):
        """
        Apply `func` to text columns based on `mode`.

        :param func: The function to apply.
        :param mode: One of "both", "left" and "right".
        :param rename: If set, use new names for results instead of replacing
            the original columns. To set `rename` in "both" mode, use a tuple
            of `str`, e.g. ("text_left_new_name", "text_right_new_name").
        :param inplace: `True` to modify inplace, `False` to return a modified
            copy. (default: `False`)
        :param verbose: Verbosity.

        Examples::
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> frame = data_pack.frame

        To apply `len` on the left text and add the result as 'length_left':
            >>> data_pack.apply_on_text(len, mode='left',
            ...                         rename='length_left',
            ...                         inplace=True,
            ...                         verbose=0)
            >>> list(frame[0].columns)
            ['id_left', 'text_left', 'length_left', 'id_right', 'text_right', \
'label']

        To do the same to the right text:
            >>> data_pack.apply_on_text(len, mode='right',
            ...                         rename='length_right',
            ...                         inplace=True,
            ...                         verbose=0)
            >>> list(frame[0].columns)
            ['id_left', 'text_left', 'length_left', 'id_right', 'text_right', \
'length_right', 'label']

        To do the same to the both texts at the same time:
            >>> data_pack.apply_on_text(len, mode='both',
            ...                         rename=('extra_left', 'extra_right'),
            ...                         inplace=True,
            ...                         verbose=0)
            >>> list(frame[0].columns)
            ['id_left', 'text_left', 'length_left', 'extra_left', 'id_right', \
'text_right', 'length_right', 'extra_right', 'label']

        To suppress outputs:
            >>> data_pack.apply_on_text(len, mode='both', verbose=0,
            ...                         inplace=True)

        """
        if mode == 'both':
            self._apply_on_text_both(func, rename, verbose=verbose)
        elif mode == 'left':
            self._apply_on_text_left(func, rename, verbose=verbose)
        elif mode == 'right':
            self._apply_on_text_right(func, rename, verbose=verbose)
        else:
            raise ValueError(f"{mode} is not a valid mode type."
                             f"Must be one of `left` `right` `both`.")

    def _apply_on_text_right(self, func, rename, verbose=1):
        name = rename or 'text_right'
        if verbose:
            tqdm.pandas(desc="Processing " + name + " with " + func.__name__)
            self._right[name] = self._right['text_right'].progress_apply(func)
        else:
            self._right[name] = self._right['text_right'].apply(func)

    def _apply_on_text_left(self, func, rename, verbose=1):
        name = rename or 'text_left'
        if verbose:
            tqdm.pandas(desc="Processing " + name + " with " + func.__name__)
            self._left[name] = self._left['text_left'].progress_apply(func)
        else:
            self._left[name] = self._left['text_left'].apply(func)

    def _apply_on_text_both(self, func, rename, verbose=1):
        left_name, right_name = rename or ('text_left', 'text_right')
        self._apply_on_text_left(func, rename=left_name, verbose=verbose)
        self._apply_on_text_right(func, rename=right_name, verbose=verbose)

    @_optional_inplace
    def one_hot_encode_label(self, num_classes=2):
        """
        One-hot encode `label` column of `relation`.

        :param num_classes: Number of classes.
        :param inplace: `True` to modify inplace, `False` to return a modified
            copy. (default: `False`)
        :return:
        """
        self._relation['label'] = self._relation['label'].apply(
            lambda idx: matchzoo.one_hot(idx, num_classes))

    class FrameView(object):
        """FrameView."""

        def __init__(self, data_pack: 'DataPack'):
            """
            View a data pack as a frame.

            A slice of the view is genereated by merging three parts of the
            data pack being viewed into a big table.

            :param data_pack: :class:`DataPack` to view.

            Examples::
                >>> import matchzoo as mz
                >>> data_pack = mz.datasets.toy.load_data()
                >>> frame = data_pack.frame

            Use `()` to get a full copy of the frame:
                >>> list(frame().columns)
                ['id_left', 'text_left', 'id_right', 'text_right', 'label']
                >>> len(frame()) == len(data_pack)
                True

            Notice that a view is binded to the original data pack, so changing
            contents of the data pack will affect a view previously created:
                >>> data_pack.drop_label(inplace=True)
                >>> list(frame().columns)
                ['id_left', 'text_left', 'id_right', 'text_right']

            To slice the view:
                >>> frame_slice = frame[3:5]
                >>> len(frame_slice)
                2

            """
            self._data_pack = data_pack

        def __getitem__(self, index: typing.Union[int, slice, np.array]
                        ) -> pd.DataFrame:
            """Slicer."""
            dp = self._data_pack
            index = _convert_to_list_index(index, len(dp))
            left_df = dp.left.loc[dp.relation['id_left'][index]].reset_index()
            right_df = dp.right.loc[
                dp.relation['id_right'][index]].reset_index()
            joined_table = left_df.join(right_df)
            for column in dp.relation.columns:
                if column not in ['id_left', 'id_right']:
                    labels = dp.relation[column][index].to_frame()
                    labels = labels.reset_index(drop=True)
                    joined_table = joined_table.join(labels)
            return joined_table

        def __call__(self):
            """:return: A full copy. Equivalant to `frame[:]`."""
            return self[:]


def load_data_pack(dirpath: typing.Union[str, Path]) -> DataPack:
    """
    Load a :class:`DataPack`. The reverse function of :meth:`save`.

    :param dirpath: directory path of the saved model.
    :return: a :class:`DataPack` instance.
    """
    dirpath = Path(dirpath)

    data_file_path = dirpath.joinpath(DataPack.DATA_FILENAME)
    dp = dill.load(open(data_file_path, 'rb'))

    return dp



"""Average precision metric for ranking."""
import numpy as np

from matchzoo.engine import base_metric
from . import Precision


class AveragePrecision(base_metric.BaseMetric):
    """Average precision metric."""

    ALIAS = ['average_precision', 'ap']

    def __init__(self, threshold: float = 0.):
        """
        :class:`AveragePrecision` constructor.

        :param threshold: The label threshold of relevance degree.
        """
        self._threshold = threshold

    def __repr__(self) -> str:
        """:return: Formated string representation of the metric."""
        return f"{self.ALIAS[0]}({self._threshold})"

    def __call__(self, y_true: np.array, y_pred: np.array) -> float:
        """
        Calculate average precision (area under PR curve).

        Example:
            >>> y_true = [0, 1]
            >>> y_pred = [0.1, 0.6]
            >>> round(AveragePrecision()(y_true, y_pred), 2)
            0.75
            >>> round(AveragePrecision()([], []), 2)
            0.0

        :param y_true: The ground true label of each document.
        :param y_pred: The predicted scores of each document.
        :return: Average precision.
        """
        precision_metrics = [Precision(k + 1) for k in range(len(y_pred))]
        out = [metric(y_true, y_pred) for metric in precision_metrics]
        if not out:
            return 0.
        return np.asscalar(np.mean(out))



import matchzoo as mz
from matchzoo.data_generator.data_generator import DataGenerator


class DataGeneratorBuilder(object):
    """
    Data Generator Bulider. In essense a wrapped partial function.

    Example:
        >>> import matchzoo as mz
        >>> builder = mz.DataGeneratorBuilder(mode='pair', batch_size=32)
        >>> data = mz.datasets.toy.load_data()
        >>> gen = builder.build(data)
        >>> type(gen)
        <class 'matchzoo.data_generator.data_generator.DataGenerator'>
        >>> gen.batch_size
        32
        >>> gen_64 = builder.build(data, batch_size=64)
        >>> gen_64.batch_size
        64

    """

    def __init__(self, **kwargs):
        """Init."""
        self._kwargs = kwargs

    def build(self, data_pack, **kwargs) -> DataGenerator:
        """
        Build a DataGenerator.

        :param data_pack: DataPack to build upon.
        :param kwargs: Additional keyword arguments to override the keyword
            arguments passed in `__init__`.
        """
        return mz.DataGenerator(data_pack, **{**self._kwargs, **kwargs})



"""Embedding data loader."""

from pathlib import Path

import keras

import matchzoo as mz

_glove_embedding_url = "http://nlp.stanford.edu/data/glove.6B.zip"


def load_glove_embedding(dimension: int = 50) -> mz.embedding.Embedding:
    """
    Return the pretrained glove embedding.

    :param dimension: the size of embedding dimension, the value can only be
        50, 100, or 300.
    :return: The :class:`mz.embedding.Embedding` object.
    """
    file_name = 'glove.6B.' + str(dimension) + 'd.txt'
    file_path = (Path(mz.USER_DATA_DIR) / 'glove').joinpath(file_name)
    if not file_path.exists():
        keras.utils.data_utils.get_file('glove_embedding',
                                        _glove_embedding_url,
                                        extract=True,
                                        cache_dir=mz.USER_DATA_DIR,
                                        cache_subdir='glove')
    return mz.embedding.load_from_file(file_path=str(file_path), mode='glove')



"""Discounted cumulative gain metric for ranking."""
import math

import numpy as np

from matchzoo.engine.base_metric import BaseMetric, sort_and_couple


class DiscountedCumulativeGain(BaseMetric):
    """Disconunted cumulative gain metric."""

    ALIAS = ['discounted_cumulative_gain', 'dcg']

    def __init__(self, k: int = 1, threshold: float = 0.):
        """
        :class:`DiscountedCumulativeGain` constructor.

        :param k: Number of results to consider.
        :param threshold: the label threshold of relevance degree.
        """
        self._k = k
        self._threshold = threshold

    def __repr__(self) -> str:
        """:return: Formated string representation of the metric."""
        return f"{self.ALIAS[0]}@{self._k}({self._threshold})"

    def __call__(self, y_true: np.array, y_pred: np.array) -> float:
        """
        Calculate discounted cumulative gain (dcg).

        Relevance is positive real values or binary values.

        Example:
            >>> y_true = [0, 1, 2, 0]
            >>> y_pred = [0.4, 0.2, 0.5, 0.7]
            >>> DiscountedCumulativeGain(1)(y_true, y_pred)
            0.0
            >>> round(DiscountedCumulativeGain(k=-1)(y_true, y_pred), 2)
            0.0
            >>> round(DiscountedCumulativeGain(k=2)(y_true, y_pred), 2)
            2.73
            >>> round(DiscountedCumulativeGain(k=3)(y_true, y_pred), 2)
            2.73
            >>> type(DiscountedCumulativeGain(k=1)(y_true, y_pred))
            <class 'float'>

        :param y_true: The ground true label of each document.
        :param y_pred: The predicted scores of each document.

        :return: Discounted cumulative gain.
        """
        if self._k <= 0:
            return 0.
        coupled_pair = sort_and_couple(y_true, y_pred)
        result = 0.
        for i, (label, score) in enumerate(coupled_pair):
            if i >= self._k:
                break
            if label > self._threshold:
                result += (math.pow(2., label) - 1.) / math.log(2. + i)
        return result



import numpy as np
from keras import backend as K

from matchzoo import losses


def test_hinge_loss():
    true_value = K.variable(np.array([[1.2], [1],
                                      [1], [1]]))
    pred_value = K.variable(np.array([[1.2], [0.1],
                                      [0], [-0.3]]))
    expected_loss = (0 + 1 - 0.3 + 0) / 2.0
    loss = K.eval(losses.RankHingeLoss()(true_value, pred_value))
    assert np.isclose(expected_loss, loss)
    expected_loss = (2 + 0.1 - 1.2 + 2 - 0.3 + 0) / 2.0
    loss = K.eval(losses.RankHingeLoss(margin=2)(true_value, pred_value))
    assert np.isclose(expected_loss, loss)
    true_value = K.variable(np.array([[1.2], [1], [0.8],
                                      [1], [1], [0.8]]))
    pred_value = K.variable(np.array([[1.2], [0.1], [-0.5],
                                      [0], [0], [-0.3]]))
    expected_loss = (0 + 1 - 0.15) / 2.0
    loss = K.eval(losses.RankHingeLoss(num_neg=2, margin=1)(
        true_value, pred_value))
    assert np.isclose(expected_loss, loss)


def test_rank_crossentropy_loss():
    losses.neg_num = 1

    def softmax(x):
        return np.exp(x) / np.sum(np.exp(x), axis=0)

    true_value = K.variable(np.array([[1.], [0.],
                                      [0.], [1.]]))
    pred_value = K.variable(np.array([[0.8], [0.1],
                                      [0.8], [0.1]]))
    expected_loss = (-np.log(softmax([0.8, 0.1])[0]) - np.log(
        softmax([0.8, 0.1])[1])) / 2
    loss = K.eval(losses.RankCrossEntropyLoss()(true_value, pred_value))
    assert np.isclose(expected_loss, loss)
    true_value = K.variable(np.array([[1.], [0.], [0.],
                                      [0.], [1.], [0.]]))
    pred_value = K.variable(np.array([[0.8], [0.1], [0.1],
                                      [0.8], [0.1], [0.1]]))
    expected_loss = (-np.log(softmax([0.8, 0.1, 0.1])[0]) - np.log(
        softmax([0.8, 0.1, 0.1])[1])) / 2
    loss = K.eval(losses.RankCrossEntropyLoss(num_neg=2)(
        true_value, pred_value))
    assert np.isclose(expected_loss, loss)



import nltk

from .unit import Unit


class StopRemoval(Unit):
    """
    Process unit to remove stop words.

    Example:
        >>> unit = StopRemoval()
        >>> unit.transform(['a', 'the', 'test'])
        ['test']
        >>> type(unit.stopwords)
        <class 'list'>
    """

    def __init__(self, lang: str = 'english'):
        """Initialization."""
        self._lang = lang
        self._stop = nltk.corpus.stopwords.words(self._lang)

    def transform(self, input_: list) -> list:
        """
        Remove stopwords from list of tokenized tokens.

        :param input_: list of tokenized tokens.
        :param lang: language code for stopwords.

        :return tokens: list of tokenized tokens without stopwords.
        """
        return [token
                for token
                in input_
                if token not in self._stop]

    @property
    def stopwords(self) -> list:
        """
        Get stopwords based on language.

        :params lang: language code.
        :return: list of stop words.
        """
        return self._stop



import abc
import typing

from .unit import Unit


class StatefulUnit(Unit, metaclass=abc.ABCMeta):
    """Process unit do persive state (i.e. need fit)."""

    def __init__(self):
        """Initialization."""
        self._state = {}

    @property
    def state(self):
        """Get current state."""
        return self._state

    @abc.abstractmethod
    def fit(self, input: typing.Any):
        """Abstract base method, need to be implemented in subclass."""



"""An implementation of MatchPyramid Model."""
import typing

import keras

import matchzoo
from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class MatchPyramid(BaseModel):
    """
    MatchPyramid Model.

    Examples:
        >>> model = MatchPyramid()
        >>> model.params['embedding_output_dim'] = 300
        >>> model.params['num_blocks'] = 2
        >>> model.params['kernel_count'] = [16, 32]
        >>> model.params['kernel_size'] = [[3, 3], [3, 3]]
        >>> model.params['dpool_size'] = [3, 10]
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(with_embedding=True)
        params.add(Param(name='num_blocks', value=1,
                         desc="Number of convolution blocks."))
        params.add(Param(name='kernel_count', value=[32],
                         desc="The kernel count of the 2D convolution "
                              "of each block."))
        params.add(Param(name='kernel_size', value=[[3, 3]],
                         desc="The kernel size of the 2D convolution "
                              "of each block."))
        params.add(Param(name='activation', value='relu',
                         desc="The activation function."))
        params.add(Param(name='dpool_size', value=[3, 10],
                         desc="The max-pooling size of each block."))
        params.add(Param(
            name='padding', value='same',
            desc="The padding mode in the convolution layer."
        ))
        params.add(Param(
            name='dropout_rate', value=0.0,
            hyper_space=hyper_spaces.quniform(low=0.0, high=0.8,
                                              q=0.01),
            desc="The dropout rate."
        ))
        return params

    def build(self):
        """
        Build model structure.

        MatchPyramid text matching as image recognition.
        """
        input_left, input_right = self._make_inputs()
        input_dpool_index = keras.layers.Input(
            name='dpool_index',
            shape=[self._params['input_shapes'][0][0],
                   self._params['input_shapes'][1][0],
                   2],
            dtype='int32')

        embedding = self._make_embedding_layer()
        embed_left = embedding(input_left)
        embed_right = embedding(input_right)

        # Interaction
        matching_layer = matchzoo.layers.MatchingLayer(matching_type='dot')
        embed_cross = matching_layer([embed_left, embed_right])

        for i in range(self._params['num_blocks']):
            embed_cross = self._conv_block(
                embed_cross,
                self._params['kernel_count'][i],
                self._params['kernel_size'][i],
                self._params['padding'],
                self._params['activation']
            )

        # Dynamic Pooling
        dpool_layer = matchzoo.layers.DynamicPoolingLayer(
            *self._params['dpool_size'])
        embed_pool = dpool_layer([embed_cross, input_dpool_index])

        embed_flat = keras.layers.Flatten()(embed_pool)
        x = keras.layers.Dropout(rate=self._params['dropout_rate'])(embed_flat)

        inputs = [input_left, input_right, input_dpool_index]
        x_out = self._make_output_layer()(x)
        self._backend = keras.Model(inputs=inputs, outputs=x_out)

    @classmethod
    def _conv_block(
        cls, x,
        kernel_count: int,
        kernel_size: int,
        padding: str,
        activation: str
    ) -> typing.Any:
        output = keras.layers.Conv2D(kernel_count,
                                     kernel_size,
                                     padding=padding,
                                     activation=activation)(x)
        return output



"""An implementation of DSSM, Deep Structured Semantic Model."""
from keras.models import Model
from keras.layers import Input, Dot

from matchzoo.engine.param_table import ParamTable
from matchzoo.engine.base_model import BaseModel
from matchzoo import preprocessors


class DSSM(BaseModel):
    """
    Deep structured semantic model.

    Examples:
        >>> model = DSSM()
        >>> model.params['mlp_num_layers'] = 3
        >>> model.params['mlp_num_units'] = 300
        >>> model.params['mlp_num_fan_out'] = 128
        >>> model.params['mlp_activation_func'] = 'relu'
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(with_multi_layer_perceptron=True)
        return params

    def build(self):
        """
        Build model structure.

        DSSM use Siamese arthitecture.
        """
        dim_triletter = self._params['input_shapes'][0][0]
        input_shape = (dim_triletter,)
        base_network = self._make_multi_layer_perceptron_layer()
        # Left input and right input.
        input_left = Input(name='text_left', shape=input_shape)
        input_right = Input(name='text_right', shape=input_shape)
        # Process left & right input.
        x = [base_network(input_left),
             base_network(input_right)]
        # Dot product with cosine similarity.
        x = Dot(axes=[1, 1], normalize=True)(x)
        x_out = self._make_output_layer()(x)
        self._backend = Model(
            inputs=[input_left, input_right],
            outputs=x_out)

    @classmethod
    def get_default_preprocessor(cls):
        """:return: Default preprocessor."""
        return preprocessors.DSSMPreprocessor()



"""Base Model."""

import abc
import typing
from pathlib import Path

import dill
import numpy as np
import keras
import keras.backend as K
import pandas as pd

import matchzoo
from matchzoo import DataGenerator
from matchzoo.engine import hyper_spaces
from matchzoo.engine.base_preprocessor import BasePreprocessor
from matchzoo.engine.base_metric import BaseMetric
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine.param import Param
from matchzoo import tasks


class BaseModel(abc.ABC):
    """
    Abstract base class of all MatchZoo models.

    MatchZoo models are wrapped over keras models, and the actual keras model
    built can be accessed by `model.backend`. `params` is a set of model
    hyper-parameters that deterministically builds a model. In other words,
    `params['model_class'](params=params)` of the same `params` always create
    models with the same structure.

    :param params: Model hyper-parameters. (default: return value from
        :meth:`get_default_params`)
    :param backend: A keras model as the model backend. Usually not passed as
        an argument.

    Example:
        >>> BaseModel()  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Can't instantiate abstract class BaseModel ...
        >>> class MyModel(BaseModel):
        ...     def build(self):
        ...         pass
        >>> isinstance(MyModel(), BaseModel)
        True

    """

    BACKEND_WEIGHTS_FILENAME = 'backend_weights.h5'
    PARAMS_FILENAME = 'params.dill'

    def __init__(
        self,
        params: typing.Optional[ParamTable] = None,
        backend: typing.Optional[keras.models.Model] = None
    ):
        """Init."""
        self._params = params or self.get_default_params()
        self._backend = backend

    @classmethod
    def get_default_params(
        cls,
        with_embedding=False,
        with_multi_layer_perceptron=False
    ) -> ParamTable:
        """
        Model default parameters.

        The common usage is to instantiate :class:`matchzoo.engine.ModelParams`
            first, then set the model specific parametrs.

        Examples:
            >>> class MyModel(BaseModel):
            ...     def build(self):
            ...         print(self._params['num_eggs'], 'eggs')
            ...         print('and', self._params['ham_type'])
            ...
            ...     @classmethod
            ...     def get_default_params(cls):
            ...         params = ParamTable()
            ...         params.add(Param('num_eggs', 512))
            ...         params.add(Param('ham_type', 'Parma Ham'))
            ...         return params
            >>> my_model = MyModel()
            >>> my_model.build()
            512 eggs
            and Parma Ham

        Notice that all parameters must be serialisable for the entire model
        to be serialisable. Therefore, it's strongly recommended to use python
        native data types to store parameters.

        :return: model parameters

        """
        params = ParamTable()
        params.add(Param(
            name='model_class', value=cls,
            desc="Model class. Used internally for save/load. "
                 "Changing this may cause unexpected behaviors."
        ))
        params.add(Param(
            name='input_shapes',
            desc="Dependent on the model and data. Should be set manually."
        ))
        params.add(Param(
            name='task',
            desc="Decides model output shape, loss, and metrics."
        ))
        params.add(Param(
            name='optimizer', value='adam',
        ))
        if with_embedding:
            params.add(Param(
                name='with_embedding', value=True,
                desc="A flag used help `auto` module. Shouldn't be changed."
            ))
            params.add(Param(
                name='embedding_input_dim',
                desc='Usually equals vocab size + 1. Should be set manually.'
            ))
            params.add(Param(
                name='embedding_output_dim',
                desc='Should be set manually.'
            ))
            params.add(Param(
                name='embedding_trainable', value=True,
                desc='`True` to enable embedding layer training, '
                     '`False` to freeze embedding parameters.'
            ))
        if with_multi_layer_perceptron:
            params.add(Param(
                name='with_multi_layer_perceptron', value=True,
                desc="A flag of whether a multiple layer perceptron is used. "
                     "Shouldn't be changed."
            ))
            params.add(Param(
                name='mlp_num_units', value=128,
                desc="Number of units in first `mlp_num_layers` layers.",
                hyper_space=hyper_spaces.quniform(8, 256, 8)
            ))
            params.add(Param(
                name='mlp_num_layers', value=3,
                desc="Number of layers of the multiple layer percetron.",
                hyper_space=hyper_spaces.quniform(1, 6)
            ))
            params.add(Param(
                name='mlp_num_fan_out', value=64,
                desc="Number of units of the layer that connects the multiple "
                     "layer percetron and the output.",
                hyper_space=hyper_spaces.quniform(4, 128, 4)
            ))
            params.add(Param(
                name='mlp_activation_func', value='relu',
                desc='Activation function used in the multiple '
                     'layer perceptron.'
            ))
        return params

    @classmethod
    def get_default_preprocessor(cls) -> BasePreprocessor:
        """
        Model default preprocessor.

        The preprocessor's transform should produce a correctly shaped data
        pack that can be used for training. Some extra configuration (e.g.
        setting `input_shapes` in :class:`matchzoo.models.DSSMModel` may be
        required on the user's end.

        :return: Default preprocessor.
        """
        return matchzoo.preprocessors.BasicPreprocessor()

    @property
    def params(self) -> ParamTable:
        """:return: model parameters."""
        return self._params

    @params.setter
    def params(self, val):
        self._params = val

    @property
    def backend(self) -> keras.models.Model:
        """:return model backend, a keras model instance."""
        if not self._backend:
            raise ValueError("Backend not found."
                             "Please build the model first.")
        else:
            return self._backend

    @abc.abstractmethod
    def build(self):
        """Build model, each subclass need to impelemnt this method."""

    def compile(self):
        """
        Compile model for training.

        Only `keras` native metrics are compiled together with backend.
        MatchZoo metrics are evaluated only through :meth:`evaluate`.
        Notice that `keras` count `loss` as one of the metrics while MatchZoo
        :class:`matchzoo.engine.BaseTask` does not.

        Examples:
            >>> from matchzoo import models
            >>> model = models.Naive()
            >>> model.guess_and_fill_missing_params(verbose=0)
            >>> model.params['task'].metrics = ['mse', 'map']
            >>> model.params['task'].metrics
            ['mse', mean_average_precision(0.0)]
            >>> model.build()
            >>> model.compile()

        """
        self._backend.compile(optimizer=self._params['optimizer'],
                              loss=self._params['task'].loss)

    def fit(
        self,
        x: typing.Union[np.ndarray, typing.List[np.ndarray], dict],
        y: np.ndarray,
        batch_size: int = 128,
        epochs: int = 1,
        verbose: int = 1,
        **kwargs
    ) -> keras.callbacks.History:
        """
        Fit the model.

        See :meth:`keras.models.Model.fit` for more details.

        :param x: input data.
        :param y: labels.
        :param batch_size: number of samples per gradient update.
        :param epochs: number of epochs to train the model.
        :param verbose: 0, 1, or 2. Verbosity mode. 0 = silent, 1 = verbose,
            2 = one log line per epoch.

        Key word arguments not listed above will be propagated to keras's fit.

        :return: A `keras.callbacks.History` instance. Its history attribute
            contains all information collected during training.
        """
        return self._backend.fit(x=x, y=y,
                                 batch_size=batch_size, epochs=epochs,
                                 verbose=verbose, **kwargs)

    def fit_generator(
        self,
        generator: matchzoo.DataGenerator,
        epochs: int = 1,
        verbose: int = 1,
        **kwargs
    ) -> keras.callbacks.History:
        """
        Fit the model with matchzoo `generator`.

        See :meth:`keras.models.Model.fit_generator` for more details.

        :param generator: A generator, an instance of
            :class:`engine.DataGenerator`.
        :param epochs: Number of epochs to train the model.
        :param verbose: 0, 1, or 2. Verbosity mode. 0 = silent, 1 = verbose,
            2 = one log line per epoch.

        :return: A `keras.callbacks.History` instance. Its history attribute
            contains all information collected during training.
        """
        return self._backend.fit_generator(
            generator=generator,
            epochs=epochs,
            verbose=verbose, **kwargs
        )

    def evaluate(
        self,
        x: typing.Dict[str, np.ndarray],
        y: np.ndarray,
        batch_size: int = 128
    ) -> typing.Dict[BaseMetric, float]:
        """
        Evaluate the model.

        :param x: Input data.
        :param y: Labels.
        :param batch_size: Number of samples when `predict` for evaluation.
            (default: 128)

        Examples::
            >>> import matchzoo as mz
            >>> data_pack = mz.datasets.toy.load_data()
            >>> preprocessor = mz.preprocessors.NaivePreprocessor()
            >>> data_pack = preprocessor.fit_transform(data_pack, verbose=0)
            >>> m = mz.models.DenseBaseline()
            >>> m.params['task'] = mz.tasks.Ranking()
            >>> m.params['task'].metrics = [
            ...     'acc', 'mse', 'mae', 'ce',
            ...     'average_precision', 'precision', 'dcg', 'ndcg',
            ...     'mean_reciprocal_rank', 'mean_average_precision', 'mrr',
            ...     'map', 'MAP',
            ...     mz.metrics.AveragePrecision(threshold=1),
            ...     mz.metrics.Precision(k=2, threshold=2),
            ...     mz.metrics.DiscountedCumulativeGain(k=2),
            ...     mz.metrics.NormalizedDiscountedCumulativeGain(
            ...         k=3, threshold=-1),
            ...     mz.metrics.MeanReciprocalRank(threshold=2),
            ...     mz.metrics.MeanAveragePrecision(threshold=3)
            ... ]
            >>> m.guess_and_fill_missing_params(verbose=0)
            >>> m.build()
            >>> m.compile()
            >>> x, y = data_pack.unpack()
            >>> evals = m.evaluate(x, y)
            >>> type(evals)
            <class 'dict'>

        """
        result = dict()
        matchzoo_metrics, keras_metrics = self._separate_metrics()
        y_pred = self.predict(x, batch_size)

        for metric in keras_metrics:
            metric_func = keras.metrics.get(metric)
            result[metric] = K.eval(K.mean(
                metric_func(K.variable(y), K.variable(y_pred))))

        if matchzoo_metrics:
            if not isinstance(self.params['task'], tasks.Ranking):
                raise ValueError("Matchzoo metrics only works on ranking.")
            for metric in matchzoo_metrics:
                result[metric] = self._eval_metric_on_data_frame(
                    metric, x['id_left'], y, y_pred)

        return result

    def evaluate_generator(
        self,
        generator: DataGenerator,
        batch_size: int = 128
    ) -> typing.Dict['BaseMetric', float]:
        """
        Evaluate the model.

        :param generator: DataGenerator to evluate.
        :param batch_size: Batch size. (default: 128)
        """
        x, y = generator[:]
        return self.evaluate(x, y, batch_size=batch_size)

    def _separate_metrics(self):
        matchzoo_metrics = []
        keras_metrics = []
        for metric in self._params['task'].metrics:
            if isinstance(metric, BaseMetric):
                matchzoo_metrics.append(metric)
            else:
                keras_metrics.append(metric)
        return matchzoo_metrics, keras_metrics

    @classmethod
    def _eval_metric_on_data_frame(
        cls,
        metric: BaseMetric,
        id_left: typing.Union[list, np.array],
        y: typing.Union[list, np.array],
        y_pred: typing.Union[list, np.array]
    ):
        eval_df = pd.DataFrame(data={
            'id': id_left,
            'true': y.squeeze(),
            'pred': y_pred.squeeze()
        })
        assert isinstance(metric, BaseMetric)
        val = eval_df.groupby(by='id').apply(
            lambda df: metric(df['true'].values, df['pred'].values)
        ).mean()
        return val

    def predict(
        self,
        x: typing.Dict[str, np.ndarray],
        batch_size=128
    ) -> np.ndarray:
        """
        Generate output predictions for the input samples.

        See :meth:`keras.models.Model.predict` for more details.

        :param x: input data
        :param batch_size: number of samples per gradient update
        :return: numpy array(s) of predictions
        """
        return self._backend.predict(x=x, batch_size=batch_size)

    def save(self, dirpath: typing.Union[str, Path]):
        """
        Save the model.

        A saved model is represented as a directory with two files. One is a
        model parameters file saved by `pickle`, and the other one is a model
        h5 file saved by `keras`.

        :param dirpath: directory path of the saved model
        """
        dirpath = Path(dirpath)
        params_path = dirpath.joinpath(self.PARAMS_FILENAME)
        weights_path = dirpath.joinpath(self.BACKEND_WEIGHTS_FILENAME)

        if not dirpath.exists():
            dirpath.mkdir(parents=True)
        else:
            raise FileExistsError(f'{dirpath} already exist, fail to save.')

        self._backend.save_weights(weights_path)
        with open(params_path, mode='wb') as params_file:
            dill.dump(self._params, params_file)

    def get_embedding_layer(
        self, name: str = 'embedding'
    ) -> keras.layers.Layer:
        """
        Get the embedding layer.

        All MatchZoo models with a single embedding layer set the embedding
        layer name to `embedding`, and this method should return that layer.

        :param name: Name of the embedding layer. (default: `embedding`)
        """
        for layer in self._backend.layers:
            if layer.name == name:
                return layer
        raise ValueError(f"Layer {name} not found. Initialize your embedding "
                         f"layer with `name='{name}'`.")

    def load_embedding_matrix(
        self,
        embedding_matrix: np.ndarray,
        name: str = 'embedding'
    ):
        """
        Load an embedding matrix.

        Load an embedding matrix into the model's embedding layer. The name
        of the embedding layer is specified by `name`. For models with only
        one embedding layer, set `name='embedding'` when creating the keras
        layer, and use the default `name` when load the matrix. For models
        with more than one embedding layers, initialize keras layer with
        different layer names, and set `name` accordingly to load a matrix
        to a chosen layer.

        :param embedding_matrix: Embedding matrix to be loaded.
        :param name: Name of the layer. (default: 'embedding')
        """
        self.get_embedding_layer(name).set_weights([embedding_matrix])

    def guess_and_fill_missing_params(self, verbose=1):
        """
        Guess and fill missing parameters in :attr:`params`.

        Use this method to automatically fill-in other hyper parameters.
        This involves some guessing so the parameter it fills could be
        wrong. For example, the default task is `Ranking`, and if we do not
        set it to `Classification` manaully for data packs prepared for
        classification, then the shape of the model output and the data will
        mismatch.

        :param verbose: Verbosity.
        """
        self._params.get('task').set_default(tasks.Ranking(), verbose)
        self._params.get('input_shapes').set_default([(30,), (30,)], verbose)
        if 'with_embedding' in self._params:
            self._params.get('embedding_input_dim').set_default(300, verbose)
            self._params.get('embedding_output_dim').set_default(300, verbose)

    def _set_param_default(self, name: str,
                           default_val: str, verbose: int = 0):
        if self._params[name] is None:
            self._params[name] = default_val
            if verbose:
                print(f"Parameter \"{name}\" set to {default_val}.")

    def _make_inputs(self) -> list:
        input_left = keras.layers.Input(
            name='text_left',
            shape=self._params['input_shapes'][0]
        )
        input_right = keras.layers.Input(
            name='text_right',
            shape=self._params['input_shapes'][1]
        )
        return [input_left, input_right]

    def _make_output_layer(self) -> keras.layers.Layer:
        """:return: a correctly shaped keras dense layer for model output."""
        task = self._params['task']
        if isinstance(task, tasks.Classification):
            return keras.layers.Dense(task.num_classes, activation='softmax')
        elif isinstance(task, tasks.Ranking):
            return keras.layers.Dense(1, activation='linear')
        else:
            raise ValueError(f"{task} is not a valid task type."
                             f"Must be in `Ranking` and `Classification`.")

    def _make_embedding_layer(self, name: str = 'embedding'
                              ) -> keras.layers.Layer:
        return keras.layers.Embedding(
            self._params['embedding_input_dim'],
            self._params['embedding_output_dim'],
            trainable=self._params['embedding_trainable'],
            name=name
        )

    def _make_multi_layer_perceptron_layer(self) -> keras.layers.Layer:
        # TODO: do not create new layers for a second call
        if not self._params['with_multi_layer_perceptron']:
            raise AttributeError(
                'Parameter `with_multi_layer_perception` not set.')

        def _wrapper(x):
            activation = self._params['mlp_activation_func']
            for _ in range(self._params['mlp_num_layers']):
                x = keras.layers.Dense(self._params['mlp_num_units'],
                                       activation=activation)(x)
            return keras.layers.Dense(self._params['mlp_num_fan_out'],
                                      activation=activation)(x)

        return _wrapper


def load_model(dirpath: typing.Union[str, Path]) -> BaseModel:
    """
    Load a model. The reverse function of :meth:`BaseModel.save`.

    :param dirpath: directory path of the saved model
    :return: a :class:`BaseModel` instance
    """
    dirpath = Path(dirpath)

    params_path = dirpath.joinpath(BaseModel.PARAMS_FILENAME)
    weights_path = dirpath.joinpath(BaseModel.BACKEND_WEIGHTS_FILENAME)

    with open(params_path, mode='rb') as params_file:
        params = dill.load(params_file)

    model_instance = params['model_class'](params=params)
    model_instance.build()
    model_instance.compile()
    model_instance.backend.load_weights(weights_path)
    return model_instance



"""Base task."""

import typing
import abc

from matchzoo.engine import base_metric
from matchzoo.engine import parse_metric


class BaseTask(abc.ABC):
    """Base Task, shouldn't be used directly."""

    def __init__(self, loss=None, metrics=None):
        """
        Base task constructor.

        :param loss: By default the first loss in available losses.
        :param metrics:
        """
        self._loss = loss
        self._metrics = self._convert_metrics(metrics)
        self._assure_loss()
        self._assure_metrics()

    def _convert_metrics(self, metrics):
        if not metrics:
            metrics = []
        elif not isinstance(metrics, list):
            metrics = [metrics]
        return [
            parse_metric.parse_metric(metric, self) for metric in metrics
        ]

    def _assure_loss(self):
        if not self._loss:
            self._loss = self.list_available_losses()[0]

    def _assure_metrics(self):
        if not self._metrics:
            first_available = self.list_available_metrics()[0]
            self._metrics = self._convert_metrics(first_available)

    @property
    def loss(self):
        """:return: Loss used in the task."""
        return self._loss

    @property
    def metrics(self):
        """:return: Metrics used in the task."""
        return self._metrics

    @metrics.setter
    def metrics(
        self,
        new_metrics: typing.Union[
            typing.List[str],
            typing.List[base_metric.BaseMetric],
            str,
            base_metric.BaseMetric
        ]
    ):
        self._metrics = self._convert_metrics(new_metrics)

    @classmethod
    @abc.abstractmethod
    def list_available_losses(cls) -> list:
        """:return: a list of available losses."""

    @classmethod
    @abc.abstractmethod
    def list_available_metrics(cls) -> list:
        """:return: a list of available metrics."""

    @property
    @abc.abstractmethod
    def output_shape(self) -> tuple:
        """:return: output shape of a single sample of the task."""

    @property
    @abc.abstractmethod
    def output_dtype(self):
        """:return: output data type for specific task."""



"""An implementation of Dynamic Pooling Layer."""
import typing

from keras import backend as K
from keras.engine import Layer


class DynamicPoolingLayer(Layer):
    """
    Layer that computes dynamic pooling of one tensor.

    :param psize1: pooling size of dimension 1
    :param psize2: pooling size of dimension 2
    :param kwargs: Standard layer keyword arguments.

    Examples:
        >>> import matchzoo as mz
        >>> layer = mz.layers.DynamicPoolingLayer(3, 2)
        >>> num_batch, left_len, right_len, num_dim = 5, 3, 2, 10
        >>> layer.build([[num_batch, left_len, right_len, num_dim],
        ...              [num_batch, left_len, right_len, 3]])

    """

    def __init__(self,
                 psize1: int,
                 psize2: int,
                 **kwargs):
        """:class:`DynamicPoolingLayer` constructor."""
        super().__init__(**kwargs)
        self._psize1 = psize1
        self._psize2 = psize2

    def build(self, input_shape: typing.List[int]):
        """
        Build the layer.

        :param input_shape: the shapes of the input tensors,
            for DynamicPoolingLayer we need tow input tensors.
        """
        super().build(input_shape)
        input_shape_one = input_shape[0]
        self._msize1 = input_shape_one[1]
        self._msize2 = input_shape_one[2]

    def call(self, inputs: list, **kwargs) -> typing.Any:
        """
        The computation logic of DynamicPoolingLayer.

        :param inputs: two input tensors.
        """
        x, dpool_index = inputs
        dpool_shape = K.tf.shape(dpool_index)
        batch_index_one = K.tf.expand_dims(
            K.tf.expand_dims(
                K.tf.range(dpool_shape[0]), axis=-1),
            axis=-1)
        batch_index = K.tf.expand_dims(
            K.tf.tile(batch_index_one, [1, self._msize1, self._msize2]),
            axis=-1)
        dpool_index_ex = K.tf.concat([batch_index, dpool_index], axis=3)
        x_expand = K.tf.gather_nd(x, dpool_index_ex)
        stride1 = self._msize1 / self._psize1
        stride2 = self._msize2 / self._psize2

        suggestion1 = self._msize1 / stride1
        suggestion2 = self._msize2 / stride2

        if suggestion1 != self._psize1 or suggestion2 != self._psize2:
            raise ValueError("DynamicPooling Layer can not "
                             "generate ({} x {}) output feature map, "
                             "please use ({} x {} instead.)"
                             .format(self._psize1, self._psize2,
                                     suggestion1, suggestion2))

        x_pool = K.tf.nn.max_pool(x_expand,
                                  [1, stride1, stride2, 1],
                                  [1, stride1, stride2, 1],
                                  "VALID")
        return x_pool

    def compute_output_shape(self, input_shape: list) -> tuple:
        """
        Calculate the layer output shape.

        :param input_shape: the shapes of the input tensors,
            for DynamicPoolingLayer we need tow input tensors.
        """
        input_shape_one = input_shape[0]
        return (None, self._psize1, self._psize2, input_shape_one[3])

    def get_config(self) -> dict:
        """Get the config dict of DynamicPoolingLayer."""
        config = {
            'psize1': self._psize1,
            'psize2': self._psize2
        }
        base_config = super(DynamicPoolingLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""Base generator."""

import math
import typing

import keras
import numpy as np
import pandas as pd

import matchzoo as mz
from matchzoo.data_generator.callbacks import Callback


class DataGenerator(keras.utils.Sequence):
    """
    Data Generator.

    Used to divide a :class:`matchzoo.DataPack` into batches. This is helpful
    for generating batch-wise features and delaying data preprocessing to the
    `fit` time.

    See `tutorials/data_handling.ipynb` for a walkthrough.

    :param data_pack: DataPack to generator data from.
    :param mode: One of "point", "pair", and "list". (default: "point")
    :param num_dup: Number of duplications per instance, only effective when
        `mode` is "pair". (default: 1)
    :param num_neg: Number of negative samples per instance, only effective
        when `mode` is "pair". (default: 1)
    :param resample: Either to resample for each epoch, only effective when
        `mode` is "pair". (default: `True`)
    :param batch_size: Batch size. (default: 128)
    :param shuffle: Either to shuffle the samples/instances. (default: `True`)
    :param callbacks: Callbacks. See `matchzoo.data_generator.callbacks` for
        more details.

    Examples::
        >>> import numpy as np
        >>> import matchzoo as mz
        >>> np.random.seed(0)
        >>> data_pack = mz.datasets.toy.load_data()
        >>> batch_size = 8

    To generate data points:
        >>> point_gen = mz.DataGenerator(
        ...     data_pack=data_pack,
        ...     batch_size=batch_size
        ... )
        >>> len(point_gen)
        13
        >>> x, y = point_gen[0]
        >>> for key, value in sorted(x.items()):
        ...     print(key, str(value)[:30])
        id_left ['Q6' 'Q17' 'Q1' 'Q13' 'Q16' '
        id_right ['D6-6' 'D17-1' 'D1-2' 'D13-3'
        text_left ['how long is the term for fed
        text_right ['See Article I and Article II

    To generate data pairs:
        >>> pair_gen = mz.DataGenerator(
        ...     data_pack=data_pack,
        ...     mode='pair',
        ...     num_dup=4,
        ...     num_neg=4,
        ...     batch_size=batch_size,
        ...     shuffle=False
        ... )
        >>> len(pair_gen)
        3
        >>> x, y = pair_gen[0]
        >>> for key, value in sorted(x.items()):
        ...     print(key, str(value)[:30])
        id_left ['Q1' 'Q1' 'Q1' 'Q1' 'Q1' 'Q1'
        id_right ['D1-3' 'D1-4' 'D1-0' 'D1-1' '
        text_left ['how are glacier caves formed
        text_right ['A glacier cave is a cave for

    To generate data lists:
        # TODO:

    """

    def __init__(
        self,
        data_pack: mz.DataPack,
        mode='point',
        num_dup: int = 1,
        num_neg: int = 1,
        resample: bool = True,
        batch_size: int = 128,
        shuffle: bool = True,
        callbacks: typing.List[Callback] = None
    ):
        """Init."""
        if callbacks is None:
            callbacks = []

        if mode not in ('point', 'pair', 'list'):
            raise ValueError(f"{mode} is not a valid mode type."
                             f"Must be one of `point`, `pair` or `list`.")

        self._mode = mode
        self._num_dup = num_dup
        self._num_neg = num_neg
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._resample = resample
        self._orig_relation = data_pack.relation
        self._callbacks = callbacks

        if mode == 'pair':
            data_pack.relation = self._reorganize_pair_wise(
                data_pack.relation,
                num_dup=num_dup,
                num_neg=num_neg
            )

        self._data_pack = data_pack
        self._batch_indices = None

        self.reset_index()

    def __getitem__(self, item: int) -> typing.Tuple[dict, np.ndarray]:
        """Get a batch from index idx.

        :param item: the index of the batch.
        """
        if isinstance(item, slice):
            indices = sum(self._batch_indices[item], [])
        else:
            indices = self._batch_indices[item]
        batch_data_pack = self._data_pack[indices]
        self._handle_callbacks_on_batch_data_pack(batch_data_pack)
        x, y = batch_data_pack.unpack()
        self._handle_callbacks_on_batch_unpacked(x, y)
        return x, y

    def __len__(self) -> int:
        """Get the total number of batches."""
        return len(self._batch_indices)

    def on_epoch_end(self):
        """Reorganize the index array while epoch is ended."""
        if self._mode == 'pair' and self._resample:
            self._data_pack.relation = self._reorganize_pair_wise(
                relation=self._orig_relation,
                num_dup=self._num_dup,
                num_neg=self._num_neg
            )
        self.reset_index()

    def reset_index(self):
        """
        Set the :attr:`index_array`.

        Here the :attr:`index_array` records the index of all the instances.
        """
        # index pool: index -> instance index
        if self._mode == 'point':
            num_instances = len(self._data_pack)
            index_pool = list(range(num_instances))
        elif self._mode == 'pair':
            index_pool = []
            step_size = self._num_neg + 1
            num_instances = int(len(self._data_pack) / step_size)
            for i in range(num_instances):
                lower = i * step_size
                upper = (i + 1) * step_size
                indices = list(range(lower, upper))
                if indices:
                    index_pool.append(indices)
        elif self._mode == 'list':
            raise NotImplementedError(
                f'{self._mode} data generator not implemented.')
        else:
            raise ValueError(f"{self._mode} is not a valid mode type"
                             f"Must be one of `point`, `pair` or `list`.")

        if self._shuffle:
            np.random.shuffle(index_pool)

        # batch_indices: index -> batch of indices
        self._batch_indices = []
        for i in range(math.ceil(num_instances / self._batch_size)):
            lower = self._batch_size * i
            upper = self._batch_size * (i + 1)
            candidates = index_pool[lower:upper]
            if self._mode == 'pair':
                candidates = sum(candidates, [])
            if candidates:
                self._batch_indices.append(candidates)

    def _handle_callbacks_on_batch_data_pack(self, batch_data_pack):
        for callback in self._callbacks:
            callback.on_batch_data_pack(batch_data_pack)

    def _handle_callbacks_on_batch_unpacked(self, x, y):
        for callback in self._callbacks:
            callback.on_batch_unpacked(x, y)

    @property
    def callbacks(self):
        """`callbacks` getter."""
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value):
        """`callbacks` setter."""
        self._callbacks = value

    @property
    def num_neg(self):
        """`num_neg` getter."""
        return self._num_neg

    @num_neg.setter
    def num_neg(self, value):
        """`num_neg` setter."""
        self._num_neg = value
        self.reset_index()

    @property
    def num_dup(self):
        """`num_dup` getter."""
        return self._num_dup

    @num_dup.setter
    def num_dup(self, value):
        """`num_dup` setter."""
        self._num_dup = value
        self.reset_index()

    @property
    def mode(self):
        """`mode` getter."""
        return self._mode

    @mode.setter
    def mode(self, value):
        """`mode` setter."""
        self._mode = value
        self.reset_index()

    @property
    def batch_size(self):
        """`batch_size` getter."""
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value):
        """`batch_size` setter."""
        self._batch_size = value
        self.reset_index()

    @property
    def shuffle(self):
        """`shuffle` getter."""
        return self._shuffle

    @shuffle.setter
    def shuffle(self, value):
        """`shuffle` setter."""
        self._shuffle = value
        self.reset_index()

    @property
    def batch_indices(self):
        """`batch_indices` getter."""
        return self._batch_indices

    @classmethod
    def _reorganize_pair_wise(
        cls,
        relation: pd.DataFrame,
        num_dup: int = 1,
        num_neg: int = 1
    ):
        """Re-organize the data pack as pair-wise format."""
        pairs = []
        groups = relation.sort_values(
            'label', ascending=False).groupby('id_left')
        for idx, group in groups:
            labels = group.label.unique()
            for label in labels[:-1]:
                pos_samples = group[group.label == label]
                pos_samples = pd.concat([pos_samples] * num_dup)
                neg_samples = group[group.label < label]
                for _, pos_sample in pos_samples.iterrows():
                    pos_sample = pd.DataFrame([pos_sample])
                    neg_sample = neg_samples.sample(num_neg, replace=True)
                    pairs.extend((pos_sample, neg_sample))
        new_relation = pd.concat(pairs, ignore_index=True)
        return new_relation



"""Parameter class."""

import inspect
import numbers
import typing

import hyperopt.pyll

from matchzoo.engine import hyper_spaces

# Both hyperopt native spaces and matchzoo proxies are valid spaces.
SpaceType = typing.Union[hyperopt.pyll.Apply, hyper_spaces.HyperoptProxy]


class Param(object):
    """
    Parameter class.

    Basic usages with a name and  value:

        >>> param = Param('my_param', 10)
        >>> param.name
        'my_param'
        >>> param.value
        10

    Use with a validator to make sure the parameter always keeps a valid
    value.

        >>> param = Param(
        ...     name='my_param',
        ...     value=5,
        ...     validator=lambda x: 0 < x < 20
        ... )
        >>> param.validator  # doctest: +ELLIPSIS
        <function <lambda> at 0x...>
        >>> param.value
        5
        >>> param.value = 10
        >>> param.value
        10
        >>> param.value = -1
        Traceback (most recent call last):
            ...
        ValueError: Validator not satifised.
        The validator's definition is as follows:
        validator=lambda x: 0 < x < 20

    Use with a hyper space. Setting up a hyper space for a parameter makes the
    parameter tunable in a :class:`matchzoo.engine.Tuner`.

        >>> from matchzoo.engine.hyper_spaces import quniform
        >>> param = Param(
        ...     name='positive_num',
        ...     value=1,
        ...     hyper_space=quniform(low=1, high=5)
        ... )
        >>> param.hyper_space  # doctest: +ELLIPSIS
        <matchzoo.engine.hyper_spaces.quniform object at ...>
        >>> from hyperopt.pyll.stochastic import sample
        >>> hyperopt_space = param.hyper_space.convert(param.name)
        >>> samples = [sample(hyperopt_space) for _ in range(64)]
        >>> set(samples) == {1, 2, 3, 4, 5}
        True

    The boolean value of a :class:`Param` instance is only `True`
    when the value is not `None`. This is because some default falsy values
    like zero or an empty list are valid parameter values. In other words,
    the boolean value means to be "if the parameter value is filled".

        >>> param = Param('dropout')
        >>> if param:
        ...     print('OK')
        >>> param = Param('dropout', 0)
        >>> if param:
        ...     print('OK')
        OK

    A `_pre_assignment_hook` is initialized as a data type convertor if the
    value is set as a number to keep data type consistency of the parameter.
    This conversion supports python built-in numbers, `numpy` numbers, and
    any number that inherits :class:`numbers.Number`.

        >>> param = Param('float_param', 0.5)
        >>> param.value = 10
        >>> param.value
        10.0
        >>> type(param.value)
        <class 'float'>

    """

    def __init__(
        self,
        name: str,
        value: typing.Any = None,
        hyper_space: typing.Optional[SpaceType] = None,
        validator: typing.Optional[
            typing.Callable[[typing.Any], bool]] = None,
        desc: typing.Optional[str] = None,
    ):
        """
        Parameter constructor.

        :param name: Name of the parameter.
        :param value: Value of the parameter, `None` by default, which means
            "this parameter is not filled yet."
        :param hyper_space: Hyper space of the parameter, `None` by default.
            If set, then a :class:`matchzoo.engine.ParamTable` that has this
            parameter will include this `hyper_space` as a part of the
            parameter table's search space.
        :param validator: Validator of the parameter, `None` by default. If
            validation is needed, pass a callable that, given a value, returns
            a `bool`. The definition of the validator is retrieved when the
            validation fails, so either use a function or a `lambda` that
            occupies its own line for better readability.
        """
        self._name = name
        self._desc = desc

        self._value = None
        self._hyper_space = None
        self._validator = None
        self._pre_assignment_hook = None

        self.validator = validator
        self.hyper_space = hyper_space

        if value is not None:  # bypass checking if no default
            self.value = value

    @property
    def name(self) -> str:
        """:return: Name of the parameter."""
        return self._name

    @property
    def value(self) -> typing.Any:
        """:return: Value of the parameter."""
        return self._value

    @value.setter
    def value(self, new_value: typing.Any):
        """
        Set the value of parameter to `new_value`.

        Notice that this setter validates `new_value` before assignment. As
        a result, if the validaiton fails, the value of the parameter is not
        changed.

        :param new_value: New value of the parameter to set.
        """
        if self._pre_assignment_hook:
            new_value = self._pre_assignment_hook(new_value)
        self._validate(new_value)
        self._value = new_value
        if not self._pre_assignment_hook:
            self._infer_pre_assignment_hook()

    @property
    def hyper_space(self) -> SpaceType:
        """:return: Hyper space of the parameter."""
        return self._hyper_space

    @hyper_space.setter
    def hyper_space(self, new_space: SpaceType):
        """:param new_space: New space of the parameter to set."""
        self._hyper_space = new_space

    @property
    def validator(self) -> typing.Callable[[typing.Any], bool]:
        """:return: Validator of the parameter."""
        return self._validator

    @validator.setter
    def validator(self, new_validator: typing.Callable[[typing.Any], bool]):
        """:param new_validator: New space of the parameter to set."""
        if new_validator and not callable(new_validator):
            raise TypeError("Validator must be a callable or None.")
        self._validator = new_validator

    @property
    def desc(self) -> str:
        """:return: Parameter description."""
        return self._desc

    @desc.setter
    def desc(self, value: str):
        """:param value: New description of the parameter."""
        self._desc = value

    def _infer_pre_assignment_hook(self):
        if isinstance(self._value, numbers.Number):
            self._pre_assignment_hook = lambda x: type(self._value)(x)

    def _validate(self, value):
        if self._validator:
            valid = self._validator(value)
            if not valid:
                error_msg = "Validator not satifised.\n"
                error_msg += "The validator's definition is as follows:\n"
                error_msg += inspect.getsource(self._validator).strip()
                raise ValueError(error_msg)

    def __bool__(self):
        """:return: `False` when the value is `None`, `True` otherwise."""
        return self._value is not None

    def set_default(self, val, verbose=1):
        """
        Set default value, has no effect if already has a value.

        :param val: Default value to set.
        :param verbose: Verbosity.
        """
        if self._value is None:
            self.value = val
            if verbose:
                print(f"Parameter \"{self._name}\" set to {val}.")

    def reset(self):
        """
        Set the parameter's value to `None`, which means "not set".

        This method bypasses validator.

        Example:
            >>> import matchzoo as mz
            >>> param = mz.Param(
            ...     name='str', validator=lambda x: isinstance(x, str))
            >>> param.value = 'hello'
            >>> param.value = None
            Traceback (most recent call last):
                ...
            ValueError: Validator not satifised.
            The validator's definition is as follows:
            name='str', validator=lambda x: isinstance(x, str))
            >>> param.reset()
            >>> param.value is None
            True

        """
        self._value = None



"""Match LSTM model."""
import keras
import keras.backend as K

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine import hyper_spaces


class MatchLSTM(BaseModel):
    """
    Match LSTM model.

    Examples:
        >>> model = MatchLSTM()
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.params['embedding_input_dim'] = 10000
        >>> model.params['embedding_output_dim'] = 100
        >>> model.params['embedding_trainable'] = True
        >>> model.params['fc_num_units'] = 200
        >>> model.params['lstm_num_units'] = 256
        >>> model.params['dropout_rate'] = 0.5
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls):
        """Get default parameters."""
        params = super().get_default_params(with_embedding=True)
        params.add(Param(
            'lstm_num_units', 256,
            hyper_space=hyper_spaces.quniform(low=128, high=384, q=32),
            desc="The hidden size in the LSTM layer."
        ))
        params.add(Param(
            'fc_num_units', 200,
            hyper_space=hyper_spaces.quniform(
                low=100, high=300, q=20),
            desc="The hidden size in the full connection layer."
        ))
        params.add(Param(
            'dropout_rate', 0.0,
            hyper_space=hyper_spaces.quniform(
                low=0.0, high=0.9, q=0.01),
            desc="The dropout rate."
        ))
        return params

    def build(self):
        """Build model."""
        input_left, input_right = self._make_inputs()
        len_left = input_left.shape[1]
        len_right = input_right.shape[1]
        embedding = self._make_embedding_layer()
        embed_left = embedding(input_left)
        embed_right = embedding(input_right)

        lstm_left = keras.layers.LSTM(self._params['lstm_num_units'],
                                      return_sequences=True,
                                      name='lstm_left')
        lstm_right = keras.layers.LSTM(self._params['lstm_num_units'],
                                       return_sequences=True,
                                       name='lstm_right')
        encoded_left = lstm_left(embed_left)
        encoded_right = lstm_right(embed_right)

        def attention(tensors):
            """Attention layer."""
            left, right = tensors
            tensor_left = K.expand_dims(left, axis=2)
            tensor_right = K.expand_dims(right, axis=1)
            tensor_left = K.repeat_elements(tensor_left, len_right, 2)
            tensor_right = K.repeat_elements(tensor_right, len_left, 1)
            tensor_merged = K.concatenate([tensor_left, tensor_right], axis=-1)
            middle_output = keras.layers.Dense(self._params['fc_num_units'],
                                               activation='tanh')(
                tensor_merged)
            attn_scores = keras.layers.Dense(1)(middle_output)
            attn_scores = K.squeeze(attn_scores, axis=3)
            exp_attn_scores = K.exp(
                attn_scores - K.max(attn_scores, axis=-1, keepdims=True))
            exp_sum = K.sum(exp_attn_scores, axis=-1, keepdims=True)
            attention_weights = exp_attn_scores / exp_sum
            return K.batch_dot(attention_weights, right)

        attn_layer = keras.layers.Lambda(attention)
        left_attn_vec = attn_layer([encoded_left, encoded_right])
        concat = keras.layers.Concatenate(axis=1)(
            [left_attn_vec, encoded_right])
        lstm_merge = keras.layers.LSTM(self._params['lstm_num_units'] * 2,
                                       return_sequences=False,
                                       name='lstm_merge')
        merged = lstm_merge(concat)
        dropout = keras.layers.Dropout(
            rate=self._params['dropout_rate'])(merged)

        phi = keras.layers.Dense(self._params['fc_num_units'],
                                 activation='tanh')(dropout)
        inputs = [input_left, input_right]
        out = self._make_output_layer()(phi)
        self._backend = keras.Model(inputs=inputs, outputs=[out])



import numpy as np

import matchzoo as mz


class Callback(object):
    """
    DataGenerator callback base class.

    To build your own callbacks, inherit `mz.data_generator.callbacks.Callback`
    and overrides corresponding methods.

    A batch is processed in the following way:

    - slice data pack based on batch index
    - handle `on_batch_data_pack` callbacks
    - unpack data pack into x, y
    - handle `on_batch_x_y` callbacks
    - return x, y

    """

    def on_batch_data_pack(self, data_pack: mz.DataPack):
        """
        `on_batch_data_pack`.

        :param data_pack: a sliced DataPack before unpacking.
        """

    def on_batch_unpacked(self, x: dict, y: np.ndarray):
        """
        `on_batch_unpacked`.

        :param x: unpacked x.
        :param y: unpacked y.
        """



"""matchzoo/models/README.md generater."""

from pathlib import Path

import tabulate
import inspect
import pandas as pd

import matchzoo


def _generate():
    full = _make_title()
    for model_class in matchzoo.models.list_available():
        full += _make_model_class_subtitle(model_class)
        full += _make_doc_section_subsubtitle()
        full += _make_model_doc(model_class)
        model = model_class()
        full += _make_params_section_subsubtitle()
        full += _make_model_params_table(model)
    _write_to_files(full)


def _make_title():
    title = 'MatchZoo Model Reference'
    line = '*' * len(title)
    return line + '\n' + title + '\n' + line + '\n\n'


def _make_model_class_subtitle(model_class):
    subtitle = model_class.__name__
    line = '#' * len(subtitle)
    return subtitle + '\n' + line + '\n\n'


def _make_doc_section_subsubtitle():
    subsubtitle = 'Model Documentation'
    line = '*' * len(subsubtitle)
    return subsubtitle + '\n' + line + '\n\n'


def _make_params_section_subsubtitle():
    subsubtitle = 'Model Hyper Parameters'
    line = '*' * len(subsubtitle)
    return subsubtitle + '\n' + line + '\n\n'


def _make_model_doc(model_class):
    return inspect.getdoc(model_class) + '\n\n'


def _make_model_params_table(model):
    params = model.get_default_params()
    df = params.to_frame()
    df = df.rename({
        'Value': 'Default Value',
        'Hyper-Space': 'Default Hyper-Space'
    }, axis='columns')
    return tabulate.tabulate(df, tablefmt='rst', headers='keys') + '\n\n'


def _write_to_files(full):
    readme_file_path = Path(__file__).parent.joinpath('README.rst')
    doc_file_path = Path(__file__).parent.parent.parent. \
        joinpath('docs').joinpath('source').joinpath('model_reference.rst')
    for file_path in readme_file_path, doc_file_path:
        with open(file_path, 'w', encoding='utf-8') as out_file:
            out_file.write(full)


if __name__ == '__main__':
    _generate()



import matchzoo
from matchzoo.engine.base_model import BaseModel


class Callback(object):
    """
    Tuner callback base class.

    To build your own callbacks, inherit `mz.auto.tuner.callbacks.Callback`
    and overrides corresponding methods.

    A run proceeds in the following way:

    - run start (callback)
    - build model
    - build end (callback)
    - fit and evaluate model
    - collect result
    - run end (callback)

    This process is repeated for `num_runs` times in a tuner.

    """

    def on_run_start(self, tuner: 'matchzoo.auto.Tuner', sample: dict):
        """
        Callback on run start stage.

        :param tuner: Tuner.
        :param sample: Sampled hyper space. Changes to this dictionary affects
            the model building process of the tuner.
        """

    def on_build_end(self, tuner: 'matchzoo.auto.Tuner', model: BaseModel):
        """
        Callback on build end stage.

        :param tuner: Tuner.
        :param model: A built model ready for fitting and evluating. Changes
            to this model affect the fitting and evaluating process.
        """

    def on_run_end(self, tuner: 'matchzoo.auto.Tuner', model: BaseModel,
                   result: dict):
        """
        Callback on run end stage.

        :param tuner: Tuner.
        :param model: A built model done fitting and evaluating. Changes to
            the model will no longer affect the result.
        :param result: Result of the run. Changes to this dictionary will be
            visible in the return value of the `tune` method.
        """



"""Quora Question Pairs data loader."""

import typing
from pathlib import Path

import keras
import pandas as pd

import matchzoo

_url = "https://firebasestorage.googleapis.com/v0/b/mtl-sentence" \
       "-representations.appspot.com/o/data%2FQQP.zip?alt=media&" \
       "token=700c6acf-160d-4d89-81d1-de4191d02cb5"


def load_data(
    stage: str = 'train',
    task: str = 'classification',
    return_classes: bool = False,
) -> typing.Union[matchzoo.DataPack, tuple]:
    """
    Load QuoraQP data.

    :param path: `None` for download from quora, specific path for
        downloaded data.
    :param stage: One of `train`, `dev`, and `test`.
    :param task: Could be one of `ranking`, `classification` or a
        :class:`matchzoo.engine.BaseTask` instance.
    :param return_classes: Whether return classes for classification task.
    :return: A DataPack if `ranking`, a tuple of (DataPack, classes) if
        `classification`.
    """
    if stage not in ('train', 'dev', 'test'):
        raise ValueError(f"{stage} is not a valid stage."
                         f"Must be one of `train`, `dev`, and `test`.")

    data_root = _download_data()
    file_path = data_root.joinpath(f"{stage}.tsv")
    data_pack = _read_data(file_path, stage)

    if task == 'ranking':
        task = matchzoo.tasks.Ranking()
    elif task == 'classification':
        task = matchzoo.tasks.Classification()

    if isinstance(task, matchzoo.tasks.Ranking):
        return data_pack
    elif isinstance(task, matchzoo.tasks.Classification):
        if stage != 'test':
            data_pack.one_hot_encode_label(num_classes=2, inplace=True)
        if return_classes:
            return data_pack, [False, True]
        else:
            return data_pack
    else:
        raise ValueError(f"{task} is not a valid task.")


def _download_data():
    ref_path = keras.utils.data_utils.get_file(
        'quora_qp', _url, extract=True,
        cache_dir=matchzoo.USER_DATA_DIR,
        cache_subdir='quora_qp'
    )
    return Path(ref_path).parent.joinpath('QQP')


def _read_data(path, stage):
    data = pd.read_csv(path, sep='\t', error_bad_lines=False)
    data = data.dropna(axis=0, how='any').reset_index(drop=True)
    if stage in ['train', 'dev']:
        df = pd.DataFrame({
            'id_left': data['qid1'],
            'id_right': data['qid2'],
            'text_left': data['question1'],
            'text_right': data['question2'],
            'label': data['is_duplicate'].astype(int)
        })
    else:
        df = pd.DataFrame({
            'text_left': data['question1'],
            'text_right': data['question2']
        })
    return matchzoo.pack(df)
